/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_DEVICE_H
#define ATEMU_DEVICE_H

#include "defs.h"

/* Sleep modes */
#define SLEEPMODE_IDLE       0
#define SLEEPMODE_ADCNOISE   1
#define SLEEPMODE_POWERDOWN  2
#define SLEEPMODE_POWERSAVE  3
#define SLEEPMODE_RESERVED1  4
#define SLEEPMODE_RESERVED2  5
#define SLEEPMODE_STANDBY    6
#define SLEEPMODE_EXTSTANDBY 7

struct clock_obj;
typedef struct device_obj{
    /* Must have: */
    struct device_obj *next;
    void *dlhandle;
    void (*destroy)(void *dev);

    /* Sleep-related callback */
    error_t (*sleep)(void *dev, uint8_t sleepmode);
    error_t (*wake)(void *dev);

    /* Values maintained by the io manager */
    char *type;
    char *name;

    /* device specific data */
    void *dev;

    /* information for which cpu it's attached to */
    struct atmega_chip *chip;
    void (*device_event)(struct device_obj *dev, char *format, ...);
    void (*device_debug)(struct device_obj *dev, int level, char *format, ...);
    error_t (*add_clock)(struct device_obj *device, struct clock_obj* clock);
    error_t (*remove_clock)(struct device_obj *device, struct clock_obj* clock);
} device_obj;

#define device_setclock(device, clock, _tick) \
    { (device).add_clock(&(device), &(clock)); (clock).tick = _tick; }
#define device_clrclock(device, clock) \
    { (device).remove_clock(&(device), &(clock)); (clock).tick = NULL; }

typedef struct register_obj{
    /* read should return the data in normal integer form.
     * the io manager will shift the bits before OR-ing all of the
     * reads together.  Similarly, write will receive its bits
     * right-shifted as per its mask. */
    uint8_t (*read)(void *dev);
    void (*write)(void *dev, uint8_t value);
    uint8_t mask; /* AND mask for which bits apply to this write */
    uint8_t maskshift; /* how much to shift mask.  see above */
    void *dev;

    struct register_obj *next;
} register_obj;

typedef struct clock_obj{
    /* in a clocked device, called ever CPU tick */
    int (*tick)(void *dev);
    void *dev;

    struct clock_obj *next;
} clock_obj;

typedef struct interrupt_obj{
    /* called when jumping to the interrupt handler */
    int (*clear)(void *dev);
    uint8_t vector; /* which interrupt number it makes.  NOT a mask */
    void *dev;

    /* Chip data structure for associated proc */
    struct atmega_chip *chip;
    /* Callbacks into the IO Manager to handle interrupt signalling */
    error_t (*set_interrupt)(struct interrupt_obj *interrupt);
    error_t (*clear_interrupt)(struct interrupt_obj *interrupt);
    /* Vector mask */
    uint64_t vector_mask;
    struct interrupt_obj *next;
} interrupt_obj;

/* on DEVICE then NAME: */
device_obj* create_device(char *object_name);
typedef device_obj* (ioman_create_func)(char*);
/* on CONNECT: */
register_obj* add_register(void *dev, char *reg);
typedef register_obj* (ioman_register_func)(void*, char*);
/* on INT: */
interrupt_obj* add_interrupt(void *dev , char *name);
typedef interrupt_obj* (ioman_interrupt_func)(void*, char*);
/* on CLOCKED: */
clock_obj* add_clock(void *dev);
typedef clock_obj* (ioman_clock_func)(void*);
/* on PARAM: */
error_t set_param(void *dev, char *key, char *value);
typedef error_t (ioman_param_func)(void*, char *key, char *value);
/* on USES: */
error_t set_device(void *dev, char *name, device_obj *otherdev);
typedef error_t (ioman_uses_func)(void*, char *name, void *otherdev);

/*
 * Pin Driving Constants
 *
 * If a pin is driven with GET_FLAG(value, PIN_FLOAT) == 0, consider this
 * as a large current source and if the pin is configured as an output, this
 * is a logical failure.  Any thing with GET_FLAG(value, PIN_FLOAT) == 1
 * can be driven to an output as this only represents a resistor between rail
 * at best.  If floating, the constants PIN_DRIVE_LOW and PIN_DRIVE_HGIH can
 * be OR'ed in to represent a pull up or pull down.  If neither of these
 * constants are present, this represents a true floating state and a read
 * is considered undefined in the real world (as so it can be in the
 * emulator).
 *
 * On reads, PIN_DRIVE_LOW and PIN_DRIVE_HIGH can be used as flags to see
 * if the value should be considered as high or low.
 */
typedef enum pin_drive_enum {
    /* Digital logic driving an input low */
    PIN_DRIVE_LOW = 0x01,
    /* Digital logic driving an input high */
    PIN_DRIVE_HIGH = 0x02,

    /* The pin is being released by external circuitry but no pull-up exists
     * and the pin should be considered as in a High Impedance state (reads
     * are considered undefined) */
    PIN_FLOAT = 0x20,
    /* The pin is being released by external circuitry, but an external pull
     * down will make reads appear low */
    PIN_FLOAT_LOW = 0x21,
    /* The pin is being released by external circuitry, but an externall pull
     * up will make reads appear high */
    PIN_FLOAT_HIGH = 0x22
} pin_drive_enum;

/* If a device has pins that can be outputs, it should set up an output
 * callback.  This allows an external device to receive notifications of
 * changes to this pin.  Devices should only call this callback when the
 * value of the pin actually changes.
 *
 * If a pin can be both input and output, external hardware should catch
 * this so that it can reenable pullups and pulldowns when the output
 * floats.
 */
typedef error_t (*pin_drive_callback)(void *pvoid, void *device,
        pin_drive_enum value);

#endif /* ATEMU_DEVICE_H */
