/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_TIMER8_H
#define ATEMU_TIMER8_H

#include "../device.h"

/* All the control register bits */
#define TIMER8_CONTROL_FOC0   (uint8_t)0x80
#define TIMER8_CONTROL_WGM00  (uint8_t)0x40
#define TIMER8_CONTROL_WGM01  (uint8_t)0x08
#define TIMER8_CONTROL_COM01  (uint8_t)0x20
#define TIMER8_CONTROL_COM00  (uint8_t)0x10
#define TIMER8_CONTROL_CS02   (uint8_t)0x04
#define TIMER8_CONTROL_CS01   (uint8_t)0x02
#define TIMER8_CONTROL_CS00   (uint8_t)0x01
/* Waveform Generation Mode bitmask and values */
#define TIMER8_CONTROL_WGM0   (TIMER8_CONTROL_WGM00|TIMER8_CONTROL_WGM01)
#define TIMER8_CONTROL_WGM0_MODE0 0
#define TIMER8_CONTROL_WGM0_MODE1 TIMER8_CONTROL_WGM00
#define TIMER8_CONTROL_WGM0_MODE2 TIMER8_CONTROL_WGM01
#define TIMER8_CONTROL_WGM0_MODE3 TIMER8_CONTROL_WGM0
#define TIMER8_CONTROL_WGM0_PWM   TIMER8_CONTROL_WGM00
/* Compare Match Output Mode bitmask and values */
#define TIMER8_CONTROL_COM0   (TIMER8_CONTROL_COM00|TIMER8_CONTROL_COM01)
#define TIMER8_CONTROL_COM0_MODE0 0
#define TIMER8_CONTROL_COM0_MODE1 TIMER8_CONTROL_COM00
#define TIMER8_CONTROL_COM0_MODE2 TIMER8_CONTROL_COM01
#define TIMER8_CONTROL_COM0_MODE3 TIMER8_CONTROL_COM0
/* Clock Select bitmask */
#define TIMER8_CONTROL_CS0    (uint8_t)0x07

/* All the Asychronous Status Register bits */
#define TIMER8_ASSR_AS0       (uint8_t)0x08
#define TIMER8_ASSR_TCN0UB    (uint8_t)0x04
#define TIMER8_ASSR_OCR0UB    (uint8_t)0x02
#define TIMER8_ASSR_TCR0UB    (uint8_t)0x01

/* All the (used) Timer Interrupt Mask bits */
#define TIMER8_TIMSK_OCIE0    (uint8_t)0x02
#define TIMER8_TIMSK_TOIE0    (uint8_t)0x01

/* All the Timer Interrupt Flag bits */
#define TIMER8_TIFR_OCF0      (uint8_t)0x02
#define TIMER8_TIFR_TOV0      (uint8_t)0x01

/* Various 8-bit timer related constants */
#define TIMER8_BOTTOM         (uint8_t)0x00
#define TIMER8_MAX            (uint8_t)0xFF

typedef struct {
    device_obj    device;

    register_obj  reg_count;
    register_obj  reg_compare;
    register_obj  reg_control;
    register_obj  reg_intmask;
    register_obj  reg_intflag;
    register_obj  bit_tsm;
    register_obj  bit_psr;

    register_obj  reg_async;

    register_obj  bit_port;
    register_obj  bit_ddr;

    clock_obj     clk_tick;
    interrupt_obj int_compare;
    interrupt_obj int_overflow;

    /* Parameters */
    uint16_t async_divisor;

    /* Clock prescalers */
    const uint16_t *prescalers;

    /* CPU count for dividing as the external asynchronous clock */
    uint16_t async_divide;
    /* Whatever the actual clock source is, this is used for dividing */
    uint16_t clock_divide;
    /* The current value of this clock */
    uint8_t  count;
    /* The current comparision value of this clock */
    uint8_t  compare;
    /* Control flags */
    uint8_t  control;
    /* If count is being double buffered, writes to count are first written
     * to this variable and then copied into count. */
    uint8_t  count_buffer;
    uint8_t  compare_buffer;
    uint8_t  control_buffer;
    /* Asynchronous mode has this hideous triple buffering scheme */
    uint8_t  count_asyncbuffer;
    uint8_t  compare_asyncbuffer;
    uint8_t  control_asyncbuffer;

    /* Various flags */
        /* TRUE if counting down */
    uint8_t direction;
        /* TRUE if the comparision is to be skipped next tick */
    uint8_t skipnextcompare;
        /* TRUE if compare is double-buffered (PWM modes) */
    uint8_t buffercompare;
        /* TRUE if compare interrupts are enabled */
    uint8_t maskcompare;
        /* TRUE if overflow interrupts are enabled */
    uint8_t maskoverflow;
        /* TRUE if a compare interrupt has been signalled */
    uint8_t signalcompare;
        /* TRUE if an overflow interrupt has been signalled */
    uint8_t signaloverflow;
        /* State of Compare trigger at last tick */
    uint8_t lastcompare;
        /* State of Overflow trigger at last tick */
    uint8_t lastoverflow;
        /* TRUE if Timer/Counter Synchronization Mode enabled */
    uint8_t enabletsm;
        /* TRUE if Prescaler being held in reset */
    uint8_t enablepsr;
        /* State of last tick from the prescaler */
    uint8_t lasttick;
        /* Current divisor of the prescaler */
    uint16_t prescaler_divisor;

        /* TRUE if the CPU has entered a sleep mode where this timer is
         * disabled. */
    uint8_t sleeping;
        /* TRUE if the CPU is in a mode with the asynchronous clock disabled */
    uint8_t asyncsleeping;

        /* TRUE in asynchronous mode */
    uint8_t async;
        /* TRUE if a count has been buffered and needs to be copied */
    uint8_t bufferedcount;
        /* TRUE if a count has been buffered in the async triple buffer */
    uint8_t asyncbufferedcount;

        /* TRUE if in timer2 mode with the external clock source */
    uint8_t externalclock;
        /* Current status of Port D pin 7 (binary 1 or 0) */
    uint8_t externalvalue;
        /* Value of Port D pin 7 at last clock */
    uint8_t externalprevious;
} timer8_device;

/* The external clock generation pin on timer2 is handled by external
 * hardware (or possibly software driven).  In any case, it is left up to
 * external hardware to drive this pin as appropriate. */
error_t timer8_drive_clock(timer8_device* device, pin_drive_enum value);
/* typedef, for type safety encouragement in dependant modules */
typedef error_t (*timer8_drive_func)(timer8_device*, pin_drive_enum);

#endif /* ATEMU_TIMER8_H */
