/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "atemu.h"
#include "breakpoints.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "gutter.h"
#include "interface.h"

/* are the auto breakpoints set? */
int auto_breaks_active;

/* functions for using breakpoints */
void add_breakpoint(int chipid, long addr){
    char asmlabel[32], *clabel = NULL;
    char *cfile;
    GtkTreeIter iter;
    GtkWidget *view = NULL;
    asm_break_tuple *to_add;
    long int asmline, cline;
    
    chip_array[chipid].num_breaks++; 

    asmline = address_to_text_line( chipid, addr );
    
    /* make a pretty asm addr */
    snprintf( asmlabel, 32, "0x%08lx", addr );
    asmlabel[31] = '\0';

    /* make a pretty c line label */
    cline = address_to_c_line( chipid, addr, &cfile );
    if(cfile != NULL){
        int labelsize;
        /* enough room for string, 6 digits, : and terminating \0 */
        labelsize = strlen(cfile) + 6 + 2;
        clabel = malloc(labelsize);
        snprintf( clabel, labelsize,  "%s:%ld", cfile, cline + 1 );
        clabel[labelsize - 1] = '\0';
        view = cfile_to_textview(chipid, cfile);
    }

    /* add it to the text list */
    gtk_list_store_append( chip_array[chipid].breakp, &iter );
    gtk_list_store_set( GTK_LIST_STORE(chip_array[chipid].breakp), &iter,
        BREAK_ADDR_TEXT, asmlabel, BREAK_LINE, clabel, BREAK_ADDR, addr, -1 );

    /* install it onto the chip */
    atemu_add_break(chipid, addr);
    
    /* put it in the big table */
    to_add = malloc(sizeof(asm_break_tuple));
    to_add->addr = addr;
    table_store(chip_array[chipid].asm_break, to_add);

    /* and also add it to the gutter */
    gutter_draw_stock(GTK_TEXT_VIEW(chip_array[chipid].textview_code), GUTTER_COL_BREAK, asmline, GTK_STOCK_NO);
    if(view != NULL){
        gutter_draw_stock(GTK_TEXT_VIEW(view), GUTTER_COL_BREAK, cline, GTK_STOCK_NO);
    }
    
    if(clabel != NULL){
        free(clabel);
    }

    /* manage sensitivity of bp remove menu item */    
    if( curr_chip == chipid && chip_array[curr_chip].num_breaks > 0 )
    {
        gtk_widget_set_sensitive(menu_debug_clear_breaks, TRUE);
    }
}

void remove_breakpoint_by_addr(int chipid, long addr){
    asm_break_tuple query;
    GtkTreeIter iter;
    GtkWidget *view = NULL;
    char *cfile = NULL;
    long int line, cline;
    long addrinnode;
    gboolean iterexists;
    
    query.addr = addr;
    iterexists = gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );

    line = address_to_text_line( chipid, addr );
   
    /* remove it from the long list if same addr */
    while( iterexists )
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter, 
                BREAK_ADDR, &addrinnode, -1 );
        if( addrinnode == addr )
        {
            gtk_list_store_remove( GTK_LIST_STORE(chip_array[chipid].breakp), &iter );
            break;
        }
        iterexists = gtk_tree_model_iter_next( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );
    }

    cline = address_to_c_line( chipid, addr, &cfile );
    if(cfile != NULL){
        view = cfile_to_textview(chipid, cfile);
    }

    table_delete(chip_array[chipid].asm_break, &query, free);
    atemu_clear_break(chipid, addr);
    chip_array[chipid].num_breaks--; 

    /* and also remove it from the gutter */
    gutter_clear(GTK_TEXT_VIEW(chip_array[chipid].textview_code), GUTTER_COL_BREAK, line);
    if(view != NULL && !is_breakpoint_at_c_line(chipid, cfile, cline)){
        gutter_clear(GTK_TEXT_VIEW(view), GUTTER_COL_BREAK, cline);
    }
    
    /* manage sensitivity of bp remove menu item */    
    if( curr_chip == chipid && chip_array[curr_chip].num_breaks == 0 )
    {
        gtk_widget_set_sensitive(menu_debug_clear_breaks, FALSE);
    }
}

void remove_breakpoint_by_c_line(int chipid, char *cfile, long line){
    long addr, s_addr, e_addr;
    
    c_line_to_address_range(chipid, cfile, line, &s_addr, &e_addr);
    for(addr = s_addr; addr < e_addr; addr++){
        if(is_breakpoint_at_addr(chipid, addr)){
            remove_breakpoint_by_addr(chipid, addr);
        }
    }
}

static void add_auto_breakpoint(int chipid, long addr){
    auto_break_tuple *to_add;

    to_add = malloc(sizeof(auto_break_tuple));
    to_add->addr = addr;
    
    table_store(chip_array[chipid].auto_break, to_add);
}

void remove_all_breakpoints(int chipid){
    GtkTreeIter iter;
    long addr;

    while(gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter ))
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter, 
            BREAK_ADDR, &addr, -1 );
        
        remove_breakpoint_by_addr(chipid, addr); 
    }
    gtk_list_store_clear( GTK_LIST_STORE(chip_array[chipid].breakp));
    chip_array[chipid].num_breaks = 0; 
}

/* return true if there is a breakpoint at the address */
int is_breakpoint_at_addr(int chipid, long addr){
    asm_break_tuple query;

    query.addr = addr;
    
    return (table_find(chip_array[chipid].asm_break, &query) != NULL);
}

/* return true if there is a breakpoint at the address */
int is_breakpoint_at_c_line(int chipid, char *cfile, long line){
    asm_break_tuple query;
    long addr, s_addr, e_addr;

    c_line_to_address_range(chipid, cfile, line, &s_addr, &e_addr);

    for(addr = s_addr; addr < e_addr; addr++){
        query.addr = addr;
        if(table_find(chip_array[chipid].asm_break, &query) != NULL){
            return -1;
        }
    }
    return 0;
}

/* these functions make the list of automatic breakpoints.
 * later they will be activated and deactivated for C stepping */

void generate_auto_breaks_foreach(void *a, void *data){
    struct tuple_file_addr *at = (struct tuple_file_addr *)a;
    int chipid = *(int *)data;

    if(at->fname != NULL && cfile_to_textbuffer(chipid, at->fname) != NULL){
        add_auto_breakpoint(chipid, at->addr_begin);
    }
}

void generate_auto_breakpoints(int chipid){
    table_foreach(chip_array[chipid].addr_to_c_line, generate_auto_breaks_foreach, &chipid);
}

void activate_auto_breakpoints_foreach(void *a, void *data){
    auto_break_tuple *at = (auto_break_tuple *)a;
    int chipid = *(int *)data;

    atemu_add_break(chipid, at->addr);
}

void activate_auto_breakpoints(int chipid){
    /* we only need to install them if ours aren't already installed */
    if(auto_breaks_active != chipid){
        if(auto_breaks_active != -1){
            /* there are some other chip's points */
            deactivate_auto_breakpoints();
        }
        table_foreach(chip_array[chipid].auto_break, activate_auto_breakpoints_foreach, &chipid);
        auto_breaks_active = chipid;
    }
}

void deactivate_auto_breakpoints_foreach(void *a, void *data){
    auto_break_tuple *at = (auto_break_tuple *)a;
    int chipid = *(int *)data;

    atemu_clear_break(chipid, at->addr);
}

void deactivate_auto_breakpoints(){
    if(auto_breaks_active != -1){
        table_foreach(chip_array[auto_breaks_active].auto_break, deactivate_auto_breakpoints_foreach, &auto_breaks_active);
        auto_breaks_active = -1;
    }
}

/* breakpoint comparison functions */
/* I probably don't need all three */

int asm_break_comp(const void *a, const void *b){
    asm_break_tuple *at, *bt;
    at = (asm_break_tuple*)a;
    bt = (asm_break_tuple*)b;

    return at->addr - bt->addr;
}

int source_break_comp(const void *a, const void *b){
    source_break_tuple *at, *bt;
    at = (source_break_tuple*)a;
    bt = (source_break_tuple*)b;

    return at->addr - bt->addr;
}

int auto_break_comp(const void *a, const void *b){
    auto_break_tuple *at, *bt;
    at = (auto_break_tuple*)a;
    bt = (auto_break_tuple*)b;

    return at->addr - bt->addr;
}

