/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifdef HAVE_CONFIG_H
    #include <config.h>
#endif

#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "atemu.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "breakpoints.h"
#include "watchpoints.h"
#include "toolbar.h"
#include "debug.h"

static char *openpath = NULL;
static GtkWidget *curr_text_view = NULL;

void debug_clear_all_clicked( GtkWidget *w )
{
    if( curr_chip != -1 )
    {
        remove_all_breakpoints( curr_chip );
    }
}

static void file_ok_sel( GtkWidget *w, GtkFileSelection *fs )
{
    const char *filename;
    char *filecopy;
    struct stat finfo;

    filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

    /* maintain file->open path accross mutiple opens */
    if(openpath != NULL){
        free(openpath);
    }
        
    /* if the named file is a directory, cd into it */
    if(stat(filename, &finfo) == 0){
        filecopy = strdup(filename);
        
        if(finfo.st_mode & S_IFDIR){
            /* it's a directory */
            int filelen=strlen(filecopy);
            if(filecopy[filelen-1] != '/'){
                /* gtk selector needs it to end in a /, so we add one */
                /* length of other path, +1 for '/' +1 for '\0' */
                openpath = malloc(filelen + 2);
                sprintf(openpath, "%s/", filecopy);
                free(filecopy);
            }else{
                openpath = filecopy;
            }
            
            gtk_file_selection_set_filename(fs, openpath);
            return;
        }else{
            /* it's a regular file */
            char *filepath;
            /* length of other path, +1 for '/' +1 for '\0' */
            filepath = g_path_get_dirname(filecopy);
            openpath = malloc(strlen(filepath) + 2);
            sprintf(openpath, "%s/", filepath);
            free(filecopy);
            free(filepath);
        }
    }else{
        /* we got an error statting it, do nothing */
        return;
    }

    atemu_quit();

    atemu_init( filename );
    
    gtk_widget_destroy(GTK_WIDGET(fs));
}

void
menu_file_open_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *filew;

    if(openpath == NULL){
        openpath = strdup(DATADIR"/");
    }

    filew = gtk_file_selection_new ("Please select a file.");
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filew), openpath);
    gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(filew), FALSE);

    gtk_widget_show(filew);
    /* Connect the ok_button to file_ok_sel function */
    g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
            "clicked", G_CALLBACK (file_ok_sel), (gpointer) filew);

    /* Connect the cancel_button to destroy the widget */
    g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (filew)->cancel_button),
            "clicked", G_CALLBACK (gtk_widget_destroy),
            G_OBJECT (filew));
}

void
menu_file_quit_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    atemu_quit();
    gtk_main_quit( );
}

void
menu_edit_copy_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
     if( curr_text_view != NULL ) {
         g_signal_emit_by_name( G_OBJECT( curr_text_view ), "copy-clipboard" );
     } else {
         verb_printf( TRUE, stdout, "copy null\n" );
     }
}

void on_edit_select_all_activate( GtkMenuItem *menuItem, gpointer user_data )
{
    if( curr_text_view != NULL ) {
        #if defined(_WIN32) || defined(__CYGWIN__)
            /* Do it the hard way because gtk win32 lacks support for
             * the "select-all" signal. */
            
            GtkTextBuffer *win_buff = gtk_text_view_get_buffer( GTK_TEXT_VIEW( curr_text_view ) );
            GtkTextIter start, end;
            
            gtk_text_buffer_get_iter_at_offset( win_buff, &start, 0 );
            gtk_text_buffer_get_iter_at_offset( win_buff, &end, -1 );
            
            gtk_text_buffer_move_mark_by_name( win_buff, "insert", &start );
            gtk_text_buffer_move_mark_by_name( win_buff, "selection_bound", &end );
        #else
            g_signal_emit_by_name( G_OBJECT( curr_text_view ), "select-all" );
        #endif
    } else {
        verb_printf( TRUE, stdout, "select all null\n" );
    }
}

void on_edit_deselect_all_activate( GtkMenuItem *menuItem, gpointer user_data )
{
    if( curr_text_view != NULL ) {
        GtkTextBuffer *buff = gtk_text_view_get_buffer( GTK_TEXT_VIEW( curr_text_view ) );
        GtkTextIter start;
        
        gtk_text_buffer_get_iter_at_offset( buff, &start, 0 );
        
        gtk_text_buffer_place_cursor( buff, &start );
    } else {
        verb_printf( TRUE, stdout, "deselect all null\n" );
    }
}

void on_set_watch_activate( GtkWidget * widget, gpointer data ) 
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    int id;
    struct tuple_global_addr *glob;

    select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals ) );
    gtk_tree_selection_get_selected( select, &tm, &iter );

    gtk_tree_model_get( tm, &iter, GLOBALS_ID, &id, -1 );

    glob = id_to_global_data( curr_chip, id );

    if ( glob == NULL )
    {
        verb_printf( TRUE, stdout, "on_set_watch_activate: invalid id -> glob == NULL.\n" );
    } else {
        add_watchpoint( curr_chip, glob->addr, glob );
    }
}

void on_clear_watch_activate( GtkWidget *widget, gpointer data )
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;
    
    long addr;
    int id;
    struct tuple_global_addr *glob;
    
    if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_globals ) ){
        select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals ) );
        gtk_tree_selection_get_selected( select, &tm, &iter );

        gtk_tree_model_get( tm, &iter, GLOBALS_ID, &id, -1 );
        
        glob = id_to_global_data( curr_chip, id );
        if( glob )
        {
            remove_watchpoint( curr_chip, glob->addr, id );
        }
    } else if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_watch ) ) {
        int global_id;

        select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[curr_chip].treeview_watch ) );
        gtk_tree_selection_get_selected( select, &tm, &iter );
        
        /*find the address of the selection*/
        gtk_tree_model_get( tm, &iter, WATCH_ADDR, &addr, WATCH_ID, &global_id, -1 );
        
        /*remove the watch point*/
        remove_watchpoint( curr_chip, addr, global_id ); 
    } 
    
}

void on_clear_all_watchpoints_activate( GtkWidget *widget, gpointer data ) 
{
    GtkTreeIter iter;
    long addr;

    /*loop while items still exist in the list*/
    while( gtk_tree_model_get_iter_from_string( 
                GTK_TREE_MODEL(chip_array[curr_chip].watchp) ,
                &iter, "0" ) )
    {
        int global_id;

        /* find the address of the current iterator */
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].watchp), &iter,
        WATCH_ADDR, &addr, WATCH_ID, &global_id, -1 );
        
        /* remove the watch point */
        remove_watchpoint( curr_chip, addr, global_id ); 
        
        /* sensitize menu_debug_set_watch if need be */
        treeview_globals_selection_changed( NULL, NULL );
    }
    
    return;
    
}

gboolean on_window1_configure(GtkWidget *widget, GdkEventConfigure *event){
    int vpos, hpos;
    int i;

    if(vpane_ratio < 0){
        vpane_ratio = -vpane_ratio;
    }

    if(hpane_ratio < 0){
        hpane_ratio = -hpane_ratio;
    }

    hpos = (int)(hpane_ratio * ((double)event->width));
    vpos = (int)(vpane_ratio * ((double)event->height));

    for(i=0; i<chip_array_cnt; i++){
        gtk_paned_set_position(GTK_PANED (chip_array[i].hpaned), hpos);
        gtk_paned_set_position(GTK_PANED (chip_array[i].vpaned), vpos);
    }

    return FALSE;
}

void vpane_change_position(GObject *object, GParamSpec *pspec, gpointer data){
    if(vpane_ratio >= 0){
        int x, y;
        int pos;

        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        g_object_get(object, pspec->name, &pos, NULL);

        vpane_ratio = ((double)pos) / ((double)y);
    }
}

void hpane_change_position(GObject *object, GParamSpec *pspec, gpointer data){
    if(hpane_ratio >= 0){
        int x, y;
        int pos;
//        int i;

        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        g_object_get(object, pspec->name, &pos, NULL);

        hpane_ratio = ((double)pos) / ((double)x);

//        for(i=0; i<chip_array_cnt; i++){
//            printf("%d handlers removed.\n", g_signal_handlers_disconnect_by_func(chip_array[i].hpaned, hpane_change_position, NULL));
//            g_signal_stop_emission_by_name(chip_array[i].hpaned, "notify::position");
//            gtk_paned_set_position(GTK_PANED (chip_array[i].hpaned), pos);
//            g_signal_connect(G_OBJECT(chip_array[i].hpaned), "notify::position", G_CALLBACK(hpane_change_position), NULL);
//        }
    }
}

gboolean notebook_chips_select(GtkNotebook *book, GtkNotebookPage *page, guint page_num, gpointer data){
    int x, y;
    int hpos, vpos;
    
    curr_chip = page_num - 1;
    
    if(curr_chip >= 0){
        if( chip_array[curr_chip].num_breaks > 0 )
        {
            gtk_widget_set_sensitive( menu_debug_clear_breaks, TRUE );
        }
        else
        {
            gtk_widget_set_sensitive( menu_debug_clear_breaks, FALSE );
        }
        
        if( chip_array[curr_chip].num_watches > 0 )
        {
            gtk_widget_set_sensitive( menu_debug_clear_watches, TRUE );
        }
        else
        {
            gtk_widget_set_sensitive( menu_debug_clear_watches, FALSE );
        }
        
        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        hpos = (int)(hpane_ratio * ((double)x));
        vpos = (int)(vpane_ratio * ((double)y));

        gtk_paned_set_position(GTK_PANED (chip_array[curr_chip].hpaned), hpos);
        gtk_paned_set_position(GTK_PANED (chip_array[curr_chip].vpaned), vpos);
    }
    else{
        gtk_widget_set_sensitive( menu_debug_clear_breaks, FALSE );
        gtk_widget_set_sensitive( menu_debug_clear_watches, FALSE );
    }

    /* update cstep button to match current chip */
    toolbar_cstep_sense_update();
    
    return FALSE;
}

static void code_popup_add_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    GtkTextBuffer *buff;
    GtkWidget *view = (GtkWidget*)user_data;
    char *cfile;
    int line;
    long addr;

    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
    mark = gtk_text_buffer_get_mark(buff, "mark_click");
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter);

    if(view == chip_array[curr_chip].textview_code){
        /* ASM */
        addr = text_line_to_address(curr_chip, line);
    }else{
        /* C */
        cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
        addr = c_line_to_address(curr_chip, cfile, line);
    }

    add_breakpoint(curr_chip, addr);
}

static void code_popup_clear_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    GtkTextBuffer *buff;
    GtkWidget *view = (GtkWidget*)user_data;
    char *cfile;
    int line;
    long addr;

    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
    mark = gtk_text_buffer_get_mark(buff, "mark_click");
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter);

    if(view == chip_array[curr_chip].textview_code){
        /* ASM */
        addr = text_line_to_address(curr_chip, line);
        remove_breakpoint_by_addr(curr_chip, addr);
    }else{
        /* C */
        cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
        remove_breakpoint_by_c_line(curr_chip, cfile, line);
    }
}

void watch_popup_clear_watch(GtkWidget *widget, gpointer data){
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeViewColumn *column;
    long addr;
    int global_id;

    /*find out where the cursor is and highlight the row*/
    gtk_tree_view_get_cursor( (GTK_TREE_VIEW(chip_array[curr_chip].treeview_watch)), 
        &path, &column );
    gtk_tree_view_row_activated ( GTK_TREE_VIEW( chip_array[curr_chip].treeview_watch ),
         path, column );

    /*if the iter does not exist, then return*/
    if( !gtk_tree_model_get_iter( 
                GTK_TREE_MODEL(chip_array[curr_chip].watchp) ,
                &iter, path ) )
    {
        return;
    }

    /*find the address of the selection*/
    gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].watchp), &iter,
            WATCH_ADDR, &addr, WATCH_ID, &global_id, -1 );

    /*remove the watch point*/
    remove_watchpoint( curr_chip, addr, global_id ); 
    
}

void break_popup_clear_break(GtkWidget *widget, gpointer data){
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeViewColumn *column;
    long addr;

    /*find out where the cursor is and highlight the row*/
    gtk_tree_view_get_cursor( (GTK_TREE_VIEW(chip_array[curr_chip].treeview_break)), 
        &path, &column );
    gtk_tree_view_row_activated ( GTK_TREE_VIEW( chip_array[curr_chip].treeview_break ),
         path, column );

    /*if the iter does not exist, then return*/
    if( !gtk_tree_model_get_iter( 
                GTK_TREE_MODEL(chip_array[curr_chip].breakp) ,
                &iter, path ) )
    {
        return;
    }

    /*find the type and address of the selection*/
    gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].breakp), &iter, 
            BREAK_ADDR, &addr, -1 );

    /*remove the appropriate breakpoint*/
    remove_breakpoint_by_addr( curr_chip, addr );
    
}

/* this function places a mark "mark_click" at the place of clicking */
gboolean mark_to_click(GtkTextView *view, GdkEventButton *event, gpointer data) {
    if (event->button == 3) {
        int x, y;
        GtkTextIter iter;

        gtk_text_view_window_to_buffer_coords(view, 
                GTK_TEXT_WINDOW_TEXT, 
                event->x, event->y,
                &x, &y);
        gtk_text_view_get_iter_at_location(view, &iter, x, y);
        gtk_text_buffer_move_mark_by_name(gtk_text_view_get_buffer(view), "mark_click", &iter);
    }
    return FALSE;
}

gboolean break_window_popup(GtkWidget *widget, GdkEventButton *event, gpointer data){
    GtkTreePath *path;
    
    /* try to get path to the node from event location */
    if(!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW( chip_array[curr_chip].treeview_break ), event->x, event->y, &path, NULL, NULL, NULL)){
        return FALSE;
    }

    if (event->button == 3) {
        gtk_menu_popup( GTK_MENU( chip_array[curr_chip].treeview_break_popup ), NULL, NULL,
            NULL, NULL, event->button, event->time );
    }
    
    gtk_tree_path_free( path );

    return FALSE;
}    

gboolean watch_window_popup(GtkWidget *widget, GdkEventButton *event, gpointer data){
    GtkTreePath *path;
    
    /* try to get path to the node from event location */
    if(!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW( chip_array[curr_chip].treeview_watch ), event->x, event->y, &path, NULL, NULL, NULL)){
        return FALSE;
    }
    
    if (event->button == 3) {
        gtk_menu_popup( GTK_MENU( chip_array[curr_chip].treeview_watch_popup ), NULL, NULL,
            NULL, NULL, event->button, event->time );
    }
    
    gtk_tree_path_free( path );

    return FALSE;
}    

GtkWidget *xatdb_prepend_new_menu_item( char *icon_str, char *label, GtkMenuShell *menu )
{
    GtkWidget *menu_item = xatdb_new_menu_item( icon_str, label );
    gtk_menu_shell_prepend( menu, menu_item );
    
    return menu_item;
}

GtkWidget *xatdb_new_menu_item( char *icon_str, char *label )
{
    GtkWidget *menu_item;
    GtkWidget *icon;
    
    icon = gtk_image_new_from_stock( icon_str, GTK_ICON_SIZE_MENU );
    menu_item = gtk_image_menu_item_new_with_label( label );
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM( menu_item ), icon);
    gtk_widget_show_all( menu_item );
    
    return menu_item;
}

void add_breakpoint_menu_items( GtkMenuShell *menu, GtkTextView *textview, int break_on, long addr )
{
    GtkWidget *menu_item;
    
    /* clear breakpoint */
    menu_item = xatdb_prepend_new_menu_item( XATDB_STOCK_BREAK_OFF, _("Clear Breakpoint"), menu );
    g_signal_connect( G_OBJECT( menu_item ), "activate", G_CALLBACK( code_popup_clear_break ), textview );
    if( !break_on || addr == -1 ){
        gtk_widget_set_sensitive( menu_item, FALSE );
    }

    /*  add breakpoint */
    menu_item = xatdb_prepend_new_menu_item( XATDB_STOCK_BREAK_ON, _("Add Breakpoint"), menu );
    g_signal_connect( G_OBJECT( menu_item ), "activate", G_CALLBACK( code_popup_add_break ), textview );
    if( break_on || addr == -1 ){
        gtk_widget_set_sensitive( menu_item, FALSE );
    }
}

gboolean code_window_popup(GtkTextView *textview, GtkMenu *arg1, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    GtkTextBuffer *buff;
    GtkWidget *divider;
    char *cfile;
    int line;
    long addr;
    int bpoint;

    /* calculate address to see if we have a breakpoint there or not */
    buff = gtk_text_view_get_buffer(textview);
    mark = gtk_text_buffer_get_mark(buff, "mark_click");
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter);
    
    if((GtkWidget*)textview == chip_array[curr_chip].textview_code){
        /* ASM */
        addr = text_line_to_address(curr_chip, line);
        bpoint = is_breakpoint_at_addr(curr_chip, addr);
    }else{
        /* C */
        cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
        addr = c_line_to_address(curr_chip, cfile, line);
        bpoint = is_breakpoint_at_c_line(curr_chip, cfile, line);
    }

    /* divider */
    divider = gtk_separator_menu_item_new();
    gtk_widget_show(divider);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), divider);
    gtk_widget_set_sensitive(divider, FALSE);

    add_breakpoint_menu_items( GTK_MENU_SHELL( arg1 ), textview, bpoint, addr );
    
    return FALSE;
}

/* user has selected a different C file for display */
void select_cfile(GtkWidget *widget, gpointer data){
    GtkWidget *view;
    int chip_index;
    chip_data *chip = (chip_data *)data;

    /* get textview to new cfile */
    chip_index = gtk_option_menu_get_history(GTK_OPTION_MENU(chip->cfile_select));
    
    /* check to see if it is the disassembly */
    if( chip_index == 0 ) {
        /* User selected disassembly */
        gtk_widget_hide( chip->textview_source->parent);
        
        /* set up handle to disassembly for future calls */
        chip->textview_source = chip->textview_code;
        
        gtk_widget_show( chip->textview_code->parent );
        
    } else {
        /* User selected a source file */
    
        view = cfile_index_to_textview(chip->id, chip_index);
    
        /* hide old, show new */
        /* view->parent is the scrolled window it is contained within */
        gtk_widget_hide(chip->textview_source->parent);
        gtk_widget_show(view->parent);

        /* set up handle to current cfile for future calls */
        chip->textview_source = view;
    }
}

static void global_popup_clear_watch()
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    GtkTreeView *treev = GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals );

    int id;
    struct tuple_global_addr *glob;

    select = gtk_tree_view_get_selection( treev );
    gtk_tree_selection_get_selected( select, &tm, &iter );

    gtk_tree_model_get( tm, &iter, GLOBALS_ID, &id, -1 );

    glob = id_to_global_data( curr_chip, id );
    if( glob )
    {
        remove_watchpoint( curr_chip, glob->addr, id );
    }

}

static void global_popup_add_watch()
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    GtkTreeView *treev = GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals );

    int id;
    struct tuple_global_addr *glob;

    select = gtk_tree_view_get_selection( treev );
    gtk_tree_selection_get_selected( select, &tm, &iter );

    gtk_tree_model_get( tm, &iter, GLOBALS_ID, &id, -1 );

    glob = id_to_global_data( curr_chip, id );
    if ( glob == NULL )
    {
        verb_printf( TRUE, stdout, "global_popup_add_watch: invalid id -> glob == NULL.\n" );
    } else {
        add_watchpoint( curr_chip, glob->addr, glob );
    }
}

gboolean click_globals(GtkWidget *tv, GdkEventButton *event, gpointer null)
{
    GtkTreeIter iter;
    GtkTreePath *path;
    
    int id;
    struct tuple_global_addr *glob;
    
    /* try to get path to the node from event location */
    if(!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(tv), event->x, event->y, &path, NULL, NULL, NULL)){
        return FALSE;
    }
    
    /*if the iter does not exist, then return*/
    if( !gtk_tree_model_get_iter( GTK_TREE_MODEL(chip_array[curr_chip].globals), &iter, path ) )
    {
        gtk_tree_path_free( path );
        return FALSE;
    }
    
    gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].globals), &iter, GLOBALS_ID, &id, -1 );
    glob = id_to_global_data( curr_chip, id );

    if ( glob == NULL )
    {
        verb_printf( TRUE, stdout, "click_globals: invalid id results in glob == NULL.\n" );
        return FALSE;
    }
    
    if ( event->type != GDK_2BUTTON_PRESS )
    {    
        if ( event->button == 3 )
        {
            GtkWidget *global_watch;
            GtkWidget *icon;
            GtkWidget *menu_item;
            
            global_watch = gtk_menu_new();

            /* create clear watchpoint menu item */
            menu_item = xatdb_prepend_new_menu_item( XATDB_STOCK_WATCH_OFF, _("Clear Watchpoint"), GTK_MENU_SHELL( global_watch ) );
            g_signal_connect(G_OBJECT(menu_item), "activate", G_CALLBACK(global_popup_clear_watch), NULL);
            if(!is_watchpoint( curr_chip, glob->addr )){
                gtk_widget_set_sensitive(menu_item, FALSE);
            }

            /*  create add watchpoint menu item */
            menu_item = xatdb_prepend_new_menu_item( XATDB_STOCK_WATCH_ON, _("Set Watchpoint"), GTK_MENU_SHELL( global_watch ) );
            g_signal_connect(G_OBJECT(menu_item), "activate", G_CALLBACK(global_popup_add_watch), NULL);
            if(is_watchpoint( curr_chip, glob->addr )){
                gtk_widget_set_sensitive(menu_item, FALSE);
            }
            
            gtk_menu_popup( GTK_MENU( global_watch ), NULL, NULL,
                    NULL, NULL, event->button, event->time );
        }
        
    }
    else if ( event->button == 1 )
    {
        /* Toggle the watchpoint */
        char *status;
        int global_id;
        
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].globals), &iter, GLOBALS_ICON, &status, GLOBALS_ID, &global_id, -1 );

        if( *status == '\0' )
        {
            /* set the watchpoint */
            add_watchpoint( curr_chip, glob->addr, glob );
        }
        else
        {
            /* unset the watchpoint */
            remove_watchpoint( curr_chip, glob->addr, global_id );
        }
        free(status);
    }
    gtk_tree_path_free( path );

    return FALSE;
}

gboolean click_locals(GtkWidget *tv, GdkEventButton *event, gpointer null)
{
    GtkTreeView *treev = GTK_TREE_VIEW( tv );

    if ( ( event->type == GDK_2BUTTON_PRESS ) && ( event->button == 1 ) )
    {
        GtkTreeSelection *select;
        GtkTreeModel *tm;
        GtkTreeIter iter;

        int id;
        struct local *loc;

        select = gtk_tree_view_get_selection( treev );
        gtk_tree_selection_get_selected( select, &tm, &iter );

        gtk_tree_model_get( tm, &iter, LOCALS_ID, &id, -1 );

        loc = id_to_local_data( curr_chip, id );
        if ( loc )
        {
            loc->type_style = 1 - loc->type_style;
            render_local( curr_chip, loc, &iter );
        }
        
    }

    return FALSE;
}

gboolean text_selected( GtkWidget *widget, gpointer null )
{
    curr_text_view = widget;
    
    if( GTK_WIDGET_HAS_FOCUS( widget ) ) {
        /* widget has focus */
        gtk_widget_set_sensitive( menu_edit_select_all, TRUE );
        
        /* chech for the pressence of a selection */
        if( gtk_text_buffer_get_selection_bounds( gtk_text_view_get_buffer( GTK_TEXT_VIEW( widget ) ), NULL, NULL ) ) {
            /* user has a selection */
            gtk_widget_set_sensitive( menu_edit_copy, TRUE );
            gtk_widget_set_sensitive( menu_edit_deselect_all, TRUE );
        } else {
            /* there is no selection */
            gtk_widget_set_sensitive( menu_edit_copy, FALSE );
            gtk_widget_set_sensitive( menu_edit_deselect_all, FALSE );
        }
    } else {
        /* widget does not have focus */
        
        /* disable all items */
        gtk_widget_set_sensitive( menu_edit_deselect_all, FALSE );
        gtk_widget_set_sensitive( menu_edit_select_all, FALSE );
        gtk_widget_set_sensitive( menu_edit_copy, FALSE );
    }
    
    /* pass the event to any remaining handlers */
    return FALSE;
}

gboolean treeview_globals_selection_changed( GtkTreeSelection *widget, gpointer null )
{
    if( curr_chip != -1 ) {
        if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_globals ) ) {
            GtkTreeSelection *select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals ) );
            GtkTreeIter iter;
            GtkTreeModel *model;

            if( gtk_tree_selection_get_selected( select, &model, &iter ) ) {
                char *teststring;

                gtk_tree_model_get( model, &iter, GLOBALS_ICON, &teststring, -1 );

                gtk_widget_set_sensitive( menu_debug_set_watch, *teststring == '\0' );
                gtk_widget_set_sensitive( menu_debug_clear_watch, *teststring != '\0' );

                free( teststring );
                return FALSE;
            }
        } else if ( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_watch ) ) {
            return FALSE;
        }
    }

    gtk_widget_set_sensitive( menu_debug_set_watch, FALSE );
    gtk_widget_set_sensitive( menu_debug_clear_watch, FALSE );
    return FALSE;
}

gboolean treeview_watch_selection_changed( GtkTreeSelection *widget, gpointer null )
{
    if( curr_chip != -1 ) {
        if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_watch ) ) {
            GtkTreeSelection *select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[curr_chip].treeview_watch ) );
            GtkTreeIter iter;
            GtkTreeModel *model;

            if( gtk_tree_selection_get_selected( select, &model, &iter ) ) {

                gtk_widget_set_sensitive( menu_debug_set_watch, FALSE );
                gtk_widget_set_sensitive( menu_debug_clear_watch, TRUE );

                return FALSE;
            }
        } else if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_globals ) ) {
            return FALSE;
        }
    }

    gtk_widget_set_sensitive( menu_debug_set_watch, FALSE );
    gtk_widget_set_sensitive( menu_debug_clear_watch, FALSE );
    return FALSE;
}

gboolean watch_focus_out( GtkWidget *widget, gpointer null )
{
    gtk_widget_set_sensitive( menu_debug_set_watch, FALSE );
    gtk_widget_set_sensitive( menu_debug_clear_watch, FALSE );
    
    return FALSE;
}
