/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "atemu.h"
#include "dialogs.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "support.h"
#include "license.h"
#include "version.h"
#include "debug.h"
#include "interface.h"
#include "watchpoints.h"
#include "breakpoints.h"
#include "toolbar.h"

static GtkWidget *watch_popup = NULL;
static GtkWidget *break_popup = NULL;
static GtkWidget *about_popup = NULL;

static GtkWidget *watch_chip_radio2;
static GtkWidget *break_chip_radio2;
static GtkWidget *watch_chip_radio3;
static GtkWidget *break_chip_radio3;
static GtkWidget *watch_addr_radio1;
static GtkWidget *watch_addr_radio2;
static GtkWidget *break_addr_radio1;
static GtkWidget *break_addr_radio2;
static GtkWidget *break_addr_radio3;
static GtkWidget *watch_chip_option_menu;
static GtkWidget *break_chip_option_menu;
static GtkWidget *watch_chip_program_option_menu;
//static GtkWidget *break_chip_program_option_menu;
static GtkWidget *globals_option_menu;
static GtkWidget *globals_menu;
static GtkWidget *source_line_entry;
static GtkWidget *functions_option_menu;
static GtkWidget *functions_menu;
static GtkWidget *source_file_option_menu;
static GtkWidget *source_file_menu;

void close_popups( )
{
    close_popup( NULL, about_popup );
    close_popup( NULL, watch_popup );
    close_popup( NULL, break_popup );
}

static void add_accel_key( GtkWindow *window, gchar *key, GCallback callback )
{
    GtkAccelGroup *accel = gtk_accel_group_new();
    GClosure *close = g_cclosure_new( callback, NULL, NULL );
    guint accel_key;
    GdkModifierType mod_type;
    
    gtk_accelerator_parse( key, &accel_key, &mod_type );
    gtk_accel_group_connect( accel, accel_key, mod_type, 0, close );
    
    gtk_window_add_accel_group( window, accel );
}

static gboolean dialog_close( GtkAccelGroup *accel, GObject *target, guint keyval, GdkModifierType modifier )
{
    gtk_widget_destroy( GTK_WIDGET( target ) );
    
    return TRUE;
}

GtkWidget *xatdb_message_dialog_new_error( GtkWindow *parent,
                                           const gchar *message_header,
                                           const gchar *message_body )
{
    /* A convenience function for creating error dialogs. 
     * Calls xatdb_message_dialog_new() with some defaults.
     *
     * See the comments in xatdb_message_dialog_new() for propper
     * usage of message_header and message_body
     */
    
    return xatdb_message_dialog_new( parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_CLOSE,
        message_header,
        message_body );
}

GtkWidget *xatdb_message_dialog_new( GtkWindow *parent,
                                     GtkDialogFlags flags,
                                     GtkMessageType type,
                                     GtkButtonsType buttons,
                                     const gchar *message_header,
                                     const gchar *message_body )
{
    /* Creates a dialog box that is capable of accepting formatted 
     *
     * The argument list is the same as gtk_message_dialog_new except for the last two arguments:
     * message_header: The summary of the message. It should be succinct, not going into detail.
     *                 Simply say what happened, and if possible, provide a brief suggestion about 
     *                 what the user should do.
     * message_body:   This should be a more detailed description of what happened, what the use
     *                 can do about it, and other useful information relating to the error/message.
     *                 This can be an empty string, but it is highly reccomended that useful
     *                 information be given here.
     *
     * For further details pertaining to use of the header and body, see:
     *   http://developer.gnome.org/projects/gup/hig/1.0/windows.html#dialog-boxes
     */
    
    GtkWidget *dialog;
    
    /* In versions of Gtk+ prior to 2.4, we must do a bit of dirty work
     * to have a marked up dialog
     */
    
    #if GTK_MAJOR_VERSION > 2 || GTK_MINOR_VERSION >= 4
        dialog = gtk_message_dialog_new_with_markup( 
            parent, flags, type, buttons,
            _("<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s"),
            message_header,
            message_body ); 
    #else
        dialog = gtk_message_dialog_new( 
            parent, flags, type, buttons,
            _("<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s"),
            message_header,
            message_body );
        
        gtk_label_set_use_markup( GTK_LABEL( GTK_MESSAGE_DIALOG( dialog )->label ), TRUE );
    #endif
    
    gtk_dialog_set_has_separator( GTK_DIALOG( dialog ), FALSE );
    gtk_window_set_resizable( GTK_WINDOW( dialog ), FALSE );
    gtk_window_set_position( GTK_WINDOW( dialog ), GTK_WIN_POS_CENTER );
    add_accel_key( GTK_WINDOW( dialog ), "Escape", G_CALLBACK( dialog_close ) );
    
    g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
        G_CALLBACK (gtk_widget_destroy),
        GTK_OBJECT (dialog));
    
    return dialog;
}
#if 0
//change this completely
static void fixpaths( GString left_trun, GString left_full, GString right_trun, GString right_full )
{
    /* extend the truncated paths until they are unique */

    int len;
    // this wont work right
    while( !strcmp( left_trun->str, right_trun->str ) )
    {
        len = left_trun->len;
        g_string_erase( left_trun, 0, -1 );
        g_string_append( left_trun,
            g_strrstr_len( left_full->str, left_full->len - len, "/" ) );
            
        g_string_erase( right_trun, 0, -1 );
        g_string_append( right_trun,
            g_strrstr_len( right_full->str, right_full->len - len, "/" ) );
    }
}

GSList *get_truncated_paths( GSList *paths )
{
    /* Takes a sorted (by cfile_compare) GList of unique paths and creates a
     * list of truncated, but unique paths. The returned list and its items
     * must be freed.
     * 
     * paths:       the list of paths to truncate; paths are GStrings
    */
    // maybe use a tree instead of a list
    GSList *trun_paths = NULL;
    GSList *curr;
    GSList *curr_iter;
    GPrtArray *array;
    gboolean *repeat;
    
    /* build an initial list of [0] truncated paths and [1] full paths */
    curr = paths;
    while( curr != NULL )
    {
        array = g_ptr_array_sized_new( 2 );
        g_ptr_array_add( array, g_string_new( 
            g_strrstr( (gchar *)(G_STRING( g_paths->data )->str), "/" ) ) );
        g_ptr_array_add( array, curr->data );
        
        trun_paths = g_slist_append( trun_paths, array );
        
        curr = g_slist_next( curr );
    }
    
    curr = trun_paths;
    while( curr != NULL )
    {
        curr_iter = curr;
        repeat = TRUE;
        while( curr_iter != NULL && repeat )
        {
            if( strcmp( G_STRING( g_ptr_array_index( G_PTR_ARRAY( curr->data ), 0 ) )->str,
                G_STRING( g_ptr_array_index( G_PTR_ARRAY( curr_iter->data ), 0 ) )->str ) == 0 )
            {
                //collect item into a temporary list
            }
            
            curr_iter = g_slist_next( curr );
        }
        fixpaths( curr->data, temp_list );
        // remove updated list items from list from list, add to finnished list
            
        
        curr = g_slist_next( curr );
    }
    
    // free memory
}
#endif

/* helper functions */
static void apply_error( char *data, int type, GtkWidget *popup )
{
    GtkWidget *dialog;
    GString *message_type_u;
    GString *message_type_l;
    GString *message_append;
    GString *message_header;
    GString *message_body;
    
    if( type == XATDB_WATCH_POPUP )
    {
        message_type_u = g_string_new( _("Watchpoint") );
        message_type_l = g_string_new( _("watchpoint") );
        message_append = g_string_new( "" );
    }
    else
    {
        message_type_u = g_string_new( _("Breakpoint") );
        message_type_l = g_string_new( _("breakpoint") );
        message_append = g_string_new( _(" and divisible by four") );
    }
    
    message_header = g_string_new( "" );
    g_string_append_printf( message_header,
        _("'%s' is not a valid %s address."),
        data, message_type_l->str );
        
    message_body = g_string_new( "" );
    g_string_append_printf( message_body,
        _("%s addresses are hexadecimal numbers greater than or equal to zero%s."),
        message_type_u->str, message_append->str );
    
    dialog = xatdb_message_dialog_new_error( GTK_WINDOW( popup ),
        message_header->str,
        message_body->str );

    gtk_widget_show( dialog );
    
    g_string_free( message_type_u, TRUE );
    g_string_free( message_type_l, TRUE );
    g_string_free( message_append, TRUE );
    g_string_free( message_header, TRUE );
    g_string_free( message_body, TRUE );
}

/* some helper callbacks */

static GList *apply_to( int which )
{
    /* returns a GList of integers referring to which chips should be applied to
     *
     * int which: whether to look at breakpoints or watchpoints
     */
    
    GtkWidget *chip_radio2;
    GtkWidget *chip_radio3;
    GtkWidget *chip_option_menu;
    GtkWidget *chip_program_option_menu;
    GList *ret = NULL;
    int i;
    int *val;
    
    if( which == XATDB_WATCH_POPUP )
    {
        chip_radio2 = watch_chip_radio2;
        chip_radio3 = watch_chip_radio3;
        chip_option_menu = watch_chip_option_menu;
        chip_program_option_menu = watch_chip_option_menu;
    }
    else
    {
        chip_radio2 = break_chip_radio2;
        chip_radio3 = break_chip_radio3;
        chip_option_menu = break_chip_option_menu;
        chip_program_option_menu = break_chip_option_menu;
    }
    
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( chip_radio2 ) ) )
    {
        val = malloc( sizeof( int ) );
        *val = gtk_option_menu_get_history( GTK_OPTION_MENU( chip_option_menu ) );
        ret = g_list_append( ret, val );
    }
    else if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( chip_radio3 ) ) )
    {
        /* not yet implemented */
    }
    else
    {
        for( i = 0; i < chip_array_cnt; i++ )
        {
            val = malloc( sizeof( int ) );
            *val = i;
            ret = g_list_append( ret, val );
        }
    }
    
    return ret;
}
static void apply_breakpoint (GtkWidget *widget, gpointer data){
    int line;
    int i;
    long addr;
    char *end;
    GList *curr, *chips;
    int selected_radio;
    
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( break_addr_radio1 ) ) )
    {
        selected_radio = 1;
    }
    else if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( break_addr_radio2 ) ) )
    {
        selected_radio = 2;
    }
    else
    {
        selected_radio = 3;
    }
    
    if( *(gtk_entry_get_text( GTK_ENTRY( data ) )) == '\0' )
    {
        return;
    }
    
    if( selected_radio == 1 && 
        ( ( addr = strtol( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), &end, 16 )) < 0 || *end != '\0' ) )
    {
        apply_error( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), XATDB_BREAK_POPUP, break_popup );
        return;
    }
    else if( selected_radio == 2 && 
        ( ( addr = strtol( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), &end, 10 )) < 0 || *end != '\0' ) )
    {
        //invalid line error
        return;
    }

    curr = chips = apply_to( XATDB_BREAK_POPUP );
    while( curr != NULL )
    {
        if( selected_radio == 1 )
        {
            /* handle input to address */
            i = *((int *)(curr->data));
            
            line = address_to_text_line( i, addr );
        
            if( line == -1 )            
            {
                apply_error( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), XATDB_BREAK_POPUP, break_popup );
                return;
            }
            
            add_breakpoint( i, addr );
        }
        else if( selected_radio == 2 )
        {
            
        }
        else
        {
            
        }
        
        free( curr->data );
        curr = g_list_next( curr );
    }
    g_list_free( chips );
    
    close_popup( NULL, (gpointer)break_popup );
}

static void apply_watchpoint(GtkWidget *widget, gpointer data){
    int i;
    long addr;
    struct tuple_global_addr *glob = NULL;
    char *end;
    GList *curr, *chips;
    
    if( *(gtk_entry_get_text( GTK_ENTRY( data ) )) == '\0' )
    {
        return;
    }
    
    if( ( addr = strtol( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), &end, 16 )) < 0 || *end != '\0' )
    {
        apply_error( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), XATDB_WATCH_POPUP, watch_popup );
        return;
    }
    
    curr = chips = apply_to( XATDB_WATCH_POPUP );
    while( curr != NULL )
    {
        i = *((int *)(curr->data));
        
        /* try to get a path to the associated global */
        glob = addr_to_global_data( i, addr );
        add_watchpoint( i, addr, glob );
        
        free( curr->data );
        curr = g_list_next( curr );
    }
    g_list_free( chips );

    close_popup( NULL, (gpointer)watch_popup );
}

void kill_popup( GtkObject *object, gpointer popup )
{
    *( (GtkWidget **)popup ) = NULL;
}

void close_popup( GtkButton *button, gpointer popup )
{
    if ( popup != NULL )
    {
        gtk_widget_destroy( popup );
    }
}

static void check_address_entry( GtkEditable *entry, gchar *new_text, gint new_text_length, gint *position, gpointer addr_radio )
{
    gchar *result = g_new (gchar, new_text_length);
    const gchar *entry_text = gtk_entry_get_text( GTK_ENTRY( entry ) );
    int count = 0;
    int i;
    
    /* collect valid input(0x1..f), throw out the rest */
    for( i = 0; i < new_text_length; i++ )
    {
        if( GTK_WIDGET( entry ) == source_line_entry && isdigit( (int)new_text[i] ) )
        {
            result[ count++ ] = new_text[i];
        }
        else if( GTK_WIDGET( entry ) != source_line_entry && ( isxdigit( (int)new_text[i] )
            || ( new_text[i] == 'x' && strlen( entry_text ) + i == 1 ) ) )
        {
            result[ count++ ] = new_text[i];
        }
    }
    
    /* if there was a change, put it in */
    if( count != new_text_length )
    {
        gdk_beep();
        g_signal_handlers_block_by_func( G_OBJECT( entry ),
                                     G_CALLBACK( check_address_entry ),
                                     addr_radio );
                                     
        gtk_editable_insert_text( entry, result, count, position );
        
        g_signal_handlers_unblock_by_func( G_OBJECT( entry ),
                                       G_CALLBACK( check_address_entry ),
                                       addr_radio );
        
        g_signal_stop_emission_by_name( G_OBJECT( entry ), "insert-text" );
    }
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( addr_radio ), TRUE );
}

static void sensitize_add( GtkWidget *entry, GdkEvent *event, gpointer add_but )
{
    char *ent = (char *)gtk_entry_get_text( GTK_ENTRY( entry ) );
    
    if( *ent == '\0' || !strcmp( ent, "0x" ) )
    {
        gtk_widget_set_sensitive( GTK_WIDGET( add_but ), FALSE );
    }
    else
    {
        gtk_widget_set_sensitive( GTK_WIDGET( add_but ), TRUE );
    }
}

static void radio_select( GtkOptionMenu *chip_option_menu, gpointer chip_radio )
{
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( chip_radio ), TRUE );
}

/* create the watch or breakpoint window */
void on_add_at_activate( gpointer callback_data, guint callback_action, GtkWidget *widget )
{
    /* callback_action points to an integer such that:
     * callback_action == XATDB_WATCH_POPUP -> we are dealing with watchpoints
     * callback_action == XATDB_BREAK_POPUP -> we are dealing with breakpoints
     */
    
    int i;

    GtkWidget *window;
    GtkWidget *vbox1;
    GtkWidget *chips_label = gtk_label_new( _("<b>On:</b>") );
    GtkWidget *addresses_label = gtk_label_new( _("<b>At:</b>") );
    GtkWidget *chip_table;
    GtkWidget *chip_sub_table;
    GtkWidget *addr_table;
    GtkWidget *addr_sub_table;
    GtkWidget *chip_radio1;
    GtkWidget *chip_radio2;
    GtkWidget *chip_radio3;
    GtkWidget *addr_entry;
    GtkWidget *addr_radio1;
    GtkWidget *align;
    GtkWidget *hsep = gtk_hseparator_new();
    GtkWidget *hbox1;
    GtkWidget *hbox2;
    GtkWidget *add_but;
    GtkWidget *add_image;
    GtkWidget *add_label;
    GtkRequisition beat;
    GtkWidget *cancel_but;
    
    /* this will hopefully one day be a GtkComboBox */
    GtkWidget *chip_menu;
    GtkWidget *chip_option_menu;
    
    /* this too */
    GtkWidget *chip_program_menu;
    GtkWidget *chip_program_option_menu;
    
    PangoFontDescription *courier = pango_font_description_new();
    pango_font_description_set_family( courier, "courier" );
    
    /* set window to the appropriate window (break or watch) */
    if( callback_action == XATDB_WATCH_POPUP )
    {
        window = watch_popup;
        watch_chip_option_menu = chip_option_menu = gtk_option_menu_new();
        watch_chip_program_option_menu = chip_program_option_menu = gtk_option_menu_new();
    }
    else
    {
        window = break_popup;
        break_chip_option_menu = chip_option_menu = gtk_option_menu_new();
        watch_chip_program_option_menu = chip_program_option_menu = gtk_option_menu_new();
    }
    
    /*avoid multiple popups*/
    if( window != NULL )
    {
        gtk_window_present( GTK_WINDOW( window ) );
        return;
    }
    
    if( chip_array_cnt == 0 )
    {
        verb_printf( TRUE, stdout, _("no chips\n") );
        return;
    }
    
    /* create the window, and perfrom some popup-type specific actions */
    if( callback_action == XATDB_WATCH_POPUP )
    {
        window = watch_popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
        gtk_window_set_title( GTK_WINDOW( watch_popup ), _("Add Watchpoint...") );
        g_signal_connect( G_OBJECT( watch_popup ), "destroy",
            G_CALLBACK( kill_popup ), (gpointer)&watch_popup );
        
        add_image = gtk_image_new_from_stock( XATDB_STOCK_WATCH_ON, GTK_ICON_SIZE_BUTTON );
        
        addr_radio1 = watch_addr_radio1 = gtk_radio_button_new_with_mnemonic( NULL, _("A_ddress:") );
    }
    else
    {
        window = break_popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
        gtk_window_set_title( GTK_WINDOW( break_popup ), _("Add Breakpoint...") );
        g_signal_connect( G_OBJECT( break_popup ), "destroy",
            G_CALLBACK( kill_popup ), &break_popup );
            
        add_image = gtk_image_new_from_stock( XATDB_STOCK_BREAK_ON, GTK_ICON_SIZE_BUTTON );
        
        addr_radio1 = break_addr_radio1 = gtk_radio_button_new_with_mnemonic( NULL, _("A_ddress:") );
    }
    
    add_accel_key( GTK_WINDOW( window ), "Escape", G_CALLBACK( dialog_close ) );
    add_accel_key( GTK_WINDOW( window ), "<Ctrl>W", G_CALLBACK( dialog_close ) );
    
    gtk_container_set_border_width (GTK_CONTAINER( window ), 12);
    
    /* create the vbox to hold the window contents */
    vbox1 = gtk_vbox_new( FALSE, 12 );
    
    gtk_container_add( GTK_CONTAINER( window ), vbox1 );
    
    /* add the chips label */
    gtk_label_set_use_markup( GTK_LABEL( chips_label ), TRUE );
    gtk_misc_set_alignment( GTK_MISC( chips_label ), 0, 0.5 );
    gtk_box_pack_start( GTK_BOX( vbox1 ), chips_label, FALSE, FALSE, 0 );
    
    /* add a table to hold the indented chip options */
    chip_table = gtk_table_new( 1, 2, FALSE );
    gtk_table_set_col_spacing( GTK_TABLE( chip_table ), 0, 24 );
    gtk_table_set_row_spacing( GTK_TABLE( chip_table ), 0, 4 );
    
    chip_sub_table = gtk_table_new( 3, 2, FALSE );
    gtk_table_set_row_spacings( GTK_TABLE( chip_sub_table ), 4 );
    gtk_table_set_col_spacing( GTK_TABLE( chip_sub_table ), 1, 12 );
    gtk_table_attach( GTK_TABLE( chip_table ), chip_sub_table, 1, 2, 1, 2, 0, 0, 0, 0 );
    gtk_box_pack_start( GTK_BOX( vbox1 ), chip_table, FALSE, FALSE, 0 );
    
    /* add the "All chips" radio button */
    chip_radio1 = gtk_radio_button_new_with_mnemonic( NULL, _("A_ll chips") );
    gtk_table_attach_defaults( GTK_TABLE( chip_sub_table ), chip_radio1, 0, 1, 0, 1 );
    
    /* add the chip program radio button and pulldown */
    if( callback_action == XATDB_WATCH_POPUP )
    {
        watch_chip_radio3 = chip_radio3 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( chip_radio1 ), _("Chips _running:") );
    }
    else
    {
        break_chip_radio3 = chip_radio3 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( chip_radio1 ), _("Chips _running:") );
    }
    
    chip_program_menu = gtk_menu_new();
    
    /* insert some non-functional prototype things into the menu */
    gtk_menu_shell_append( GTK_MENU_SHELL( chip_program_menu ), gtk_menu_item_new_with_label( "Program 1" ) );
    gtk_menu_shell_append( GTK_MENU_SHELL( chip_program_menu ), gtk_menu_item_new_with_label( "Program 2" ) );

    gtk_option_menu_set_menu( GTK_OPTION_MENU( chip_program_option_menu ), chip_program_menu );
    
    gtk_option_menu_set_history( GTK_OPTION_MENU( chip_program_option_menu ), curr_chip );
    
    g_signal_connect( G_OBJECT( chip_program_option_menu ), "changed",
        G_CALLBACK( radio_select ), chip_radio3 );

    gtk_table_attach_defaults( GTK_TABLE( chip_sub_table ), chip_radio3, 0, 1, 1, 2 );
    gtk_table_attach_defaults( GTK_TABLE( chip_sub_table ), chip_program_option_menu, 2, 3, 1, 2 );
    
    /* add the chip pulldown radio button */
    if( callback_action == XATDB_WATCH_POPUP )
    {
        watch_chip_radio2 = chip_radio2 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( chip_radio1 ), _("Chi_p:") );
    }
    else
    {
        break_chip_radio2 = chip_radio2 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( chip_radio1 ), _("Chi_p:") );
    }
    
    /* add the chip menu */
    /* this will hopefully one day be a GtkComboBox */
    chip_menu = gtk_menu_new();
    
    for( i = 0; i < chip_array_cnt; i++ )
    {
        // This needs to be updated once we have support for names
        /*chip = g_string_new( chip_array[chip].name );
        g_string_append_printf( chip, "%d", i );*/
        gtk_menu_shell_append( GTK_MENU_SHELL( chip_menu ), gtk_menu_item_new_with_label( chip_array[i].name ) );
    }   

    gtk_option_menu_set_menu( GTK_OPTION_MENU( chip_option_menu ), chip_menu );
    gtk_option_menu_set_history( GTK_OPTION_MENU( chip_option_menu ), curr_chip );
    
    g_signal_connect( G_OBJECT( chip_option_menu ), "changed",
        G_CALLBACK( radio_select ), chip_radio2 );
    
    /* Select the pulldown radio button if the user has a chip selected */
    if( curr_chip != -1 )
    {
        gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( chip_radio2 ), TRUE );
    }
    
    gtk_table_attach_defaults( GTK_TABLE( chip_sub_table ), chip_radio2, 0, 1, 2, 3 );
    gtk_table_attach_defaults( GTK_TABLE( chip_sub_table ), chip_option_menu, 2, 3, 2, 3 );
    
    /* add the "At" label */
    gtk_label_set_use_markup( GTK_LABEL( addresses_label ), TRUE );
    gtk_misc_set_alignment( GTK_MISC( addresses_label ), 0, 0.5 );
    gtk_box_pack_start( GTK_BOX( vbox1 ), addresses_label, FALSE, FALSE, 0 );
    
    /* add a table to hold the address related items */
    addr_table = gtk_table_new( 1, 2, FALSE );
    gtk_table_set_col_spacing( GTK_TABLE( addr_table ), 0, 24 );
    
    addr_sub_table = gtk_table_new( 5, 3, FALSE );
    gtk_table_set_row_spacing( GTK_TABLE( addr_sub_table ), 0, 4 );
    gtk_table_set_row_spacing( GTK_TABLE( addr_sub_table ), 1, 4 );
    gtk_table_set_col_spacing( GTK_TABLE( addr_sub_table ), 1, 12 );
    
    gtk_table_attach( GTK_TABLE( addr_table ), addr_sub_table, 1, 2, 0, 1, 0, 0, 0, 0 ); 
    gtk_box_pack_start( GTK_BOX( vbox1 ), addr_table, FALSE, FALSE, 0 );
    
    /* add the address input field */
    addr_entry = gtk_entry_new();
    
    gtk_entry_set_text( GTK_ENTRY( addr_entry ), "0x" );
    gtk_entry_set_max_length( GTK_ENTRY( addr_entry ), 8 );
    gtk_widget_modify_font( addr_entry, courier );
    gtk_entry_set_width_chars( GTK_ENTRY( addr_entry ), 8 );
    
    gtk_entry_set_activates_default( GTK_ENTRY( addr_entry ), TRUE );
    
    /* Make an alignment widget for the entry */
    align = gtk_alignment_new( 0, 0.5, 0, 0 );
    gtk_container_add( GTK_CONTAINER( align ), addr_entry );
    
    gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), addr_radio1, 1, 2, 0, 1 );
    gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), align, 2, 3, 0, 1 );
    
    g_signal_connect( G_OBJECT( addr_entry ), "insert-text",
            G_CALLBACK( check_address_entry ), (gpointer) addr_radio1 );
    
    /* add some items specific to breakpoints and watchpoints */
    if( callback_action == XATDB_WATCH_POPUP )
    {
        watch_addr_radio2 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( addr_radio1 ), _("_Global:") );
        
        /* Add a dropdown for globals, as per usual, it would be nice if it was a gtk_combo_box */
        globals_option_menu = gtk_option_menu_new();
        globals_menu = gtk_menu_new();
        gtk_option_menu_set_menu( GTK_OPTION_MENU( globals_option_menu ), globals_menu );
        
        g_signal_connect( G_OBJECT( globals_option_menu ), "changed",
            G_CALLBACK( radio_select ), watch_addr_radio2 );
        
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), watch_addr_radio2, 1, 2, 1, 2 );
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), globals_option_menu, 2, 3, 1, 2 );
    }
    else
    {
        GtkWidget *in_file_label = gtk_label_new_with_mnemonic( _("_In file:") );
        
        break_addr_radio2 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( addr_radio1 ), _("_Source Line:") );
        break_addr_radio3 = gtk_radio_button_new_with_mnemonic_from_widget( GTK_RADIO_BUTTON( addr_radio1 ), _("_Function:") );
        
        gtk_table_set_row_spacing( GTK_TABLE( addr_sub_table ), 2, 2 );
        gtk_table_set_row_spacing( GTK_TABLE( addr_sub_table ), 3, 2 );
        gtk_table_set_row_spacing( GTK_TABLE( addr_sub_table ), 4, 2 );
        
        /* Add an entry for source lines */
        source_line_entry = gtk_entry_new();
        gtk_widget_modify_font( source_line_entry, courier );
        gtk_entry_set_activates_default( GTK_ENTRY( addr_entry ), TRUE );
        gtk_entry_set_width_chars( GTK_ENTRY( source_line_entry ), 10 );
        gtk_entry_set_max_length( GTK_ENTRY( source_line_entry ), 10 );
        
        align = gtk_alignment_new( 0, 0.5, 0, 0 );
        gtk_container_add( GTK_CONTAINER( align ), source_line_entry );
        
        g_signal_connect( G_OBJECT( source_line_entry ), "insert_text",
            G_CALLBACK( check_address_entry ), (gpointer)break_addr_radio2 );
        
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), break_addr_radio2, 1, 2, 1, 2 );
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), align, 2, 3, 1, 2 );
        
        /* Add a dropdown for source files */
        source_file_option_menu = gtk_option_menu_new();
        source_file_menu = gtk_menu_new();
        gtk_option_menu_set_menu( GTK_OPTION_MENU( source_file_option_menu ), source_file_menu );
        
        gtk_label_set_mnemonic_widget( GTK_LABEL( in_file_label ), source_file_option_menu );
        gtk_misc_set_alignment( GTK_MISC( in_file_label ), 1, 0.5 );
        gtk_misc_set_padding( GTK_MISC( in_file_label ), 2, 0 );
        
        g_signal_connect( G_OBJECT( source_file_option_menu ), "changed",
            G_CALLBACK( radio_select ), (gpointer)break_addr_radio2 );
        
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), in_file_label, 1, 2, 2, 3 );
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), source_file_option_menu, 2, 3, 2, 3 );
        
        /* Add a dropdown for function names */
        functions_option_menu = gtk_option_menu_new();
        functions_menu = gtk_menu_new();
        gtk_option_menu_set_menu( GTK_OPTION_MENU( functions_option_menu ), functions_menu );
        
        g_signal_connect( G_OBJECT( functions_option_menu ), "changed",
            G_CALLBACK( radio_select ), break_addr_radio3 );
        
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), break_addr_radio3, 1, 2, 3, 4 );
        gtk_table_attach_defaults( GTK_TABLE( addr_sub_table ), functions_option_menu, 2, 3, 3, 4 );
    }
    
    /* add a separator */
    gtk_box_pack_start( GTK_BOX( vbox1 ), hsep, TRUE, TRUE, 0 );
    
    /* add a box to hold the buttons */
    hbox1 = gtk_hbox_new( FALSE, 6 );
    gtk_box_pack_start( GTK_BOX( vbox1 ), hbox1, FALSE, FALSE, 0 );
    
    /* add the cancel button */
    cancel_but = gtk_button_new_from_stock( GTK_STOCK_CANCEL );
    
    g_signal_connect (G_OBJECT( cancel_but ),
            "clicked", G_CALLBACK( close_popup ),
            (gpointer)window);
    
    /* add the "add" button */
    hbox2 = gtk_hbox_new( FALSE, 0 );
    
    gtk_box_pack_start( GTK_BOX( hbox2 ), add_image, FALSE, FALSE, 3 );
    
    add_label = gtk_label_new_with_mnemonic( _("_Add") );
    
    gtk_box_pack_start( GTK_BOX( hbox2 ), add_label, FALSE, FALSE, 3 );
    
    add_but = gtk_button_new();
    gtk_container_add( GTK_CONTAINER( add_but ), hbox2 );
    
    gtk_widget_set_sensitive( add_but, FALSE );
    
    /* now that we have add_but, set some callbacks */
    
    g_signal_connect( G_OBJECT( addr_entry ), "event-after",
            G_CALLBACK( sensitize_add ), (gpointer) add_but );
            
    if( callback_action == XATDB_WATCH_POPUP )
    {
        g_signal_connect( G_OBJECT( add_but ), "clicked",
            G_CALLBACK( apply_watchpoint ), (gpointer) addr_entry );
    }
    else
    {
        g_signal_connect( G_OBJECT( add_but ), "clicked",
            G_CALLBACK( apply_breakpoint ), (gpointer) addr_entry );
    }
    
    GTK_WIDGET_SET_FLAGS (add_but, GTK_CAN_DEFAULT);
    
    gtk_box_pack_end( GTK_BOX( hbox1 ), add_but, FALSE, FALSE, 0 );
    gtk_box_pack_end( GTK_BOX( hbox1 ), cancel_but, FALSE, FALSE, 0 );
    
    gtk_widget_size_request( cancel_but, &beat );
    gtk_widget_set_size_request( add_but, beat.width, -1 );
    
    /* pop up the window */
    gtk_window_set_type_hint( GTK_WINDOW( window ), GDK_WINDOW_TYPE_HINT_DIALOG );
    
    gtk_widget_show_all( window );
    
    gtk_widget_grab_focus( addr_entry );
    gtk_editable_select_region( GTK_EDITABLE( addr_entry ), 2, 2 );
    
    gtk_widget_grab_default( add_but );
}

void menu_help_about_activate( GtkMenuItem *menuitem, gpointer user_data )
{
    GtkWidget *vbox;
    GtkWidget *label;
    GtkWidget *license;
    GtkWidget *scroll;
    GtkWidget *box1, *button_close;

    GtkTextBuffer *lbuff;
    GtkWidget *license_frame;
    GtkTextIter iter;
    
    /*stop more than one window from opening*/
    if( about_popup != NULL )
    {
        gtk_window_present( GTK_WINDOW( about_popup ) );
        return;
    }
    
    about_popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
    gtk_window_set_default_size( GTK_WINDOW( about_popup ), -1, 360 );
    gtk_window_set_title( GTK_WINDOW( about_popup ), _("About xatdb") );
    gtk_container_set_border_width (GTK_CONTAINER( about_popup ), 12);
    add_accel_key( GTK_WINDOW( about_popup ), "Escape", G_CALLBACK( dialog_close ) );
    add_accel_key( GTK_WINDOW( about_popup ), "<Ctrl>W", G_CALLBACK( dialog_close ) );
    
    g_signal_connect( G_OBJECT( about_popup ), "destroy",
            G_CALLBACK( kill_popup ), (gpointer)&about_popup );
    
    vbox = gtk_vbox_new( FALSE, 5 );
    
    gtk_container_add( GTK_CONTAINER( about_popup ), vbox );
    
    label = gtk_label_new( XATDB_VERSION );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );
    label = gtk_label_new( _("By:\tJonathan Polley \n\tDion Blazakis\n\tDaniel Rusk\n\tJonathan McGee\n\tKarl Seamon\n\tManish Karir\n") );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );
    label = gtk_label_new( "http://www.isr.umd.edu/CSHCN/research/atemu/" );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );

    license_frame = gtk_frame_new( NULL );
    gtk_frame_set_shadow_type( GTK_FRAME( license_frame ), GTK_SHADOW_IN );

    scroll = gtk_scrolled_window_new( NULL, NULL );
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( scroll ),
        GTK_POLICY_NEVER, GTK_POLICY_ALWAYS );

    license = gtk_text_view_new();
    lbuff = gtk_text_view_get_buffer( GTK_TEXT_VIEW( license ) );
    gtk_text_buffer_create_tag(lbuff, "monospace", 
            "family", "courier", 
            NULL);
    gtk_text_buffer_get_start_iter(lbuff, &iter);
    gtk_text_buffer_insert_with_tags_by_name( lbuff, &iter, license_text,
        strlen( license_text ), "monospace", NULL );
    gtk_text_view_set_editable( GTK_TEXT_VIEW( license ), FALSE );

    gtk_container_add( GTK_CONTAINER( scroll ), license );
    gtk_container_add( GTK_CONTAINER( license_frame ), scroll );
    gtk_box_pack_start( GTK_BOX( vbox ), license_frame, TRUE, TRUE, 2 );
    
    box1 = gtk_hbutton_box_new( );
    
    button_close = gtk_button_new_from_stock( GTK_STOCK_OK );
    gtk_button_box_set_layout( GTK_BUTTON_BOX( box1 ), 
                                GTK_BUTTONBOX_END );
     
    gtk_box_pack_end( GTK_BOX( box1 ), button_close, FALSE, FALSE, 0 );
    /* Connect the close button to destroy the widget */
    g_signal_connect (G_OBJECT( button_close ),
            "clicked", G_CALLBACK( close_popup ),
            (gpointer)about_popup );
    gtk_box_pack_end( GTK_BOX( vbox ), box1, FALSE, FALSE, 3 );

    gtk_widget_show_all( about_popup );
}

void child_error_popup()
{
    GtkWidget *dialog;
    
    toolbar_sense_set(FALSE);
    atemu_status_bar_finish();
    
    dialog = xatdb_message_dialog_new_error( GTK_WINDOW( window1 ),
                   "The atemu child process has unexpectedly terminated.",
                   "Xatdb's console may have additional information about the error." );
                   
    gtk_widget_show( dialog );
}

