/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "func_data.h"
#include "type_data.h"
#include "debug.h"

struct func_data *create_func_data( char * name, int beg, int end )
{
    struct func_data *ret;

    ret = (struct func_data *) malloc( sizeof(struct func_data) );
    if ( ret == NULL )
    {
        return NULL;
    }

    /* Empty list */
    ret->locals = NULL;
    ret->c_lines = NULL;

    ret->name = strdup( name );
    ret->addr_begin = beg;
    ret->addr_end = end;

    return ret;
}

static void add_c_line_to_func( struct func_data * fd, struct c_line * cl )
{
    if (( fd == NULL ) || ( cl == NULL ))
    {
        verb_printf( TRUE, stdout, "Passed NULL arg to add_c_line_to_func.\n" );
        return;
    }

    fd->c_lines = g_slist_append( fd->c_lines, (gpointer) cl );
}

static void add_local_to_func( struct func_data * fd, struct local * loc )
{
    if (( fd == NULL ) || ( loc == NULL ))
    {
        verb_printf( TRUE, stdout, "Passed NULL arg to add_local_to_func.\n" );
        return;
    }

    fd->locals = g_slist_append( fd->locals, (gpointer) loc );
}

int func_data_compare( const void * a, const void * b )
{
    struct func_data *afd = (struct func_data *) a;
    struct func_data *bfd = (struct func_data *) b;

    /* see chip_data:addr_compare_range */

    if ( bfd->addr_begin > afd->addr_begin )
        return -1;

    if ( bfd->addr_end <= afd->addr_begin )
        return 1;

    return 0;    
}

void free_func_data( void * a )
{
    struct func_data *fd = (struct func_data *) a;
    GSList *curr;

    if ( fd == NULL )
        return;

    for( curr = fd->locals; curr != NULL; curr = curr->next )
        free_local( curr->data );

    g_slist_free( fd->locals );

    for( curr = fd->c_lines; curr != NULL; curr = curr->next )
        free_c_line( curr->data );

    g_slist_free( fd->c_lines );

    free( fd->name );
    free( fd );
}

static void expand_local_struct( struct func_data * fd, struct local *parent )
{
    struct type_data *td = NULL;

    if ( parent != NULL )
    {
        td = parent->td;

        if (( td != NULL ) && ( td->fields != NULL ))
        {
            GSList *curr = td->fields;
            char *noname = "(no name)";

            while ( curr != NULL )
            {
                struct field *f = (struct field *) curr->data;
                char * name = f->name;
                char * fullname = NULL;

                if ( ( name == NULL ) || ( name[0] == '\0') )
                    name = noname;

                /* 1 for the '.' and another 1 for the NULL character */
                fullname = calloc( ( strlen( parent->name ) + strlen( name ) ) + 1 + 1, 1 );
                strcpy( fullname, parent->name );
                strcat( fullname, "." );
                strcat( fullname, name );

                create_and_add_local( fd, name, parent->is_param, parent->type,
                    parent->location + ( f->bit_offset / 8 ), f->type, parent->id,
                    parent );

                free( fullname );

                curr = curr->next;
            }
        }
    }
}

static void expand_local_array( struct func_data * fd, struct local *parent )
{
    char name[32];
    int index, delta, cnt;
    int elem_size = 0;
    struct type_data * td = parent->td;

    delta = ( (td->end > td->begin) ) ? (1) : (-1);
    cnt = ( (td->end - td->begin) * delta ) + 1;

    if ( td->extra != NULL )
    {
        if ( td->extra->root != NULL )
        {
            elem_size = td->extra->root->size;
        } else {
            /* CHECK: shouldn't happen... log something? */
            elem_size = td->extra->size;
        }
    }

    for( index = 0; index < cnt; index++ )
    {
        char * fullname = NULL;

        snprintf( name, 32, "[%lld]", index * delta + td->begin );
        fullname = calloc( ( strlen( parent->name ) + strlen( name ) ) + 1, 1 );
        strcpy( fullname, parent->name );
        strcat( fullname, name );

        create_and_add_local( fd, name, parent->is_param, parent->type,
            parent->location + ( index * elem_size ), td->extra, parent->id,
            parent );

        free( fullname );
    }
}

void create_and_add_local( struct func_data *fd, char * name, int is_param, int type,
    int location, struct type_data *td, int id, struct local *parent )
{
    struct local * ret;

    ret = (struct local *) malloc( sizeof(struct local) );
    if ( ret != NULL )
    {
        ret->name = strdup( name );
        ret->is_param = is_param;
        ret->type = type;
        ret->location = location;
        ret->td = td;
        ret->id = id;
        ret->parent = parent;

        ret->type_style = 0;


        add_local_to_func( fd, ret );

        if ( td == NULL )
        {
            ret->value = calloc( 4, sizeof( uint8_t ) );
        } else {
            ret->value = calloc( td->root->size, sizeof( uint8_t ) );

            if ( td->type == STRUCT )
            {
                expand_local_struct( fd, ret );
            } else if ( td->type == ARRAY ) {
                expand_local_array( fd, ret );
            }
        }
    }
}

void free_local( struct local * loc )
{
    if ( loc ){
        free( loc->name );
        free( loc->value );
    }

    free( loc );
}

void create_and_add_c_line( struct func_data * fd, long int addr, long int line )
{
    struct c_line *ret;

    ret = (struct c_line *) malloc( sizeof(struct c_line) );
    if ( ret != NULL )
    {
        ret->addr = addr;
        ret->line = line - 1;
    }
 
    add_c_line_to_func( fd, ret );
}

void free_c_line( struct c_line * cl )
{
    free( cl );
}
