/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#include "gutter.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "breakpoints.h"

#define GUTTER_COL_WIDTH        16

/* what we need to remember for a given little icon thing */
typedef struct _gutter_pixbuf_t{
    int column;
    int line;
    GdkPixbuf *pbuf;
}gutter_pixbuf_t;

/* each textview gets a gutter data structure */
typedef struct _gutter_data_t{
    /* list of gutter_pixbuf items */
    GSList *gutter_pixbufs;
    /* number of columns allocated */
    int columns;
}gutter_data_t;

/* comparison function to remove things from the gutter */
static int gutter_pixbuf_cmp(gconstpointer p_a, gconstpointer p_b){
    gutter_pixbuf_t *a = (gutter_pixbuf_t*)p_a;
    gutter_pixbuf_t *b = (gutter_pixbuf_t*)p_b;

    if(a->column == b->column){
        return a->line - b->line;
    }else{
        return a->column - b->column;
    }
}

/* comparison function to remove things from the gutter by column*/
static int gutter_pixbuf_col_cmp(gconstpointer p_a, gconstpointer p_b){
    gutter_pixbuf_t *a = (gutter_pixbuf_t*)p_a;
    gutter_pixbuf_t *b = (gutter_pixbuf_t*)p_b;

    return a->column - b->column;
}

static void gutter_pixbuf_data_free(gpointer data, gpointer user_data){
    free(data);
}

/* free the gutter_data struct upon textview destruction */
static void gutter_data_free(gpointer data){
    gutter_data_t *gutter_data = (gutter_data_t*)data;
    
    /* free all the crap on the list */
    g_slist_foreach(gutter_data->gutter_pixbufs, gutter_pixbuf_data_free, NULL);
    g_slist_free(gutter_data->gutter_pixbufs);
    
    free(gutter_data);
}

/* actually render the pixbufs to the window */
static void gutter_render_pixbuf(gpointer data, gpointer user_data){
    GdkWindow *gdkwin;
    GdkGC *gc;
    int p_w, p_h; /* pixbuf width and height */
    int d_x, d_y; /* destination x and y */
    int b_y, b_h; /* buffer y and line height*/
    GtkTextView *view = (GtkTextView*)user_data;
    gutter_pixbuf_t *pixbuf_data = (gutter_pixbuf_t*)data;
    GtkTextIter iter;

    gtk_text_buffer_get_iter_at_line(gtk_text_view_get_buffer(view), &iter, pixbuf_data->line);
    gdkwin = gtk_text_view_get_window(view, GTK_TEXT_WINDOW_LEFT);
    gc = gdk_gc_new(GDK_DRAWABLE(gdkwin));
    p_w = gdk_pixbuf_get_width(pixbuf_data->pbuf);
    p_h = gdk_pixbuf_get_height(pixbuf_data->pbuf);
    gtk_text_view_get_line_yrange(view, &iter, &b_y, &b_h);
    gtk_text_view_buffer_to_window_coords(view, GTK_TEXT_WINDOW_LEFT, 0, b_y, &d_x, &d_y);
    d_x = GUTTER_COL_WIDTH * pixbuf_data->column;
    
    /* dither offsets should probably be non-zero... */
    gdk_pixbuf_render_to_drawable(pixbuf_data->pbuf, GDK_DRAWABLE(gdkwin), gc, 0, 0, d_x, d_y, p_w, p_h, GDK_RGB_DITHER_NORMAL, 0, 0);

    g_object_unref(gc);
}

/* when the window is refreshed, redraw the pixbufs */
gboolean gutter_redraw_event(GtkWidget *widget, GdkEventExpose *event, gpointer user_data){
    gutter_data_t *gutter_data;

    /* only return that the event was handled if it was a redraw for the
     * gutter.  Other redraws get passed on to the TextView's handler */
    if(event->window == gtk_text_view_get_window(GTK_TEXT_VIEW(widget), GTK_TEXT_WINDOW_LEFT)){
        gutter_data = (gutter_data_t *)g_object_get_data(G_OBJECT(widget), "gutter-data");
        g_slist_foreach(gutter_data->gutter_pixbufs, gutter_render_pixbuf, widget);
        return TRUE;
    }else{
        return FALSE;
    }
}

/* when the user clicks on the gutter, toggle bpoint */
gboolean gutter_button_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data){
    int x, y;
    int line;
    GtkTextIter iter;
    GtkTextView *view = GTK_TEXT_VIEW(widget);
    long int address;
    int bpoint;
    char *cfile = NULL;

    gtk_text_view_window_to_buffer_coords(view,
            GTK_TEXT_WINDOW_TEXT,
            event->x, event->y,
            &x, &y);
    
    /* only respond if the click was on the gutter */
    if(event->window == gtk_text_view_get_window(view, GTK_TEXT_WINDOW_LEFT)){
        gtk_text_view_get_iter_at_location(view, &iter, x, y);
        line = gtk_text_iter_get_line(&iter);
        
        if(widget == chip_array[curr_chip].textview_code){
            address = text_line_to_address(curr_chip, line);
            bpoint = is_breakpoint_at_addr(curr_chip, address);
        }else{
            cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
            
            address = c_line_to_address(curr_chip, cfile, line);
            bpoint = is_breakpoint_at_c_line(curr_chip, cfile, line);
        }
        
        /* toggle for left clicks */
        if( event->button == 1 && address >= 0 ){
            if(widget == chip_array[curr_chip].textview_code){
                if(bpoint){
                    remove_breakpoint_by_addr(curr_chip, address);
                }else{
                    add_breakpoint(curr_chip, address);
                }
            }else{
                if(bpoint){
                    remove_breakpoint_by_c_line(curr_chip, cfile, line);
                }else{
                    add_breakpoint(curr_chip, address);
                }
            }
        }
        /* popup menu for right clicks */
        else if( event->button == 3 )
        {
            GtkWidget *popup_menu = gtk_menu_new();
            add_breakpoint_menu_items( GTK_MENU_SHELL( popup_menu ), view, bpoint, address );
            
            gtk_menu_popup( GTK_MENU( popup_menu ), NULL, NULL,
                    NULL, NULL, event->button, event->time );
        }
    }
    
    /* allow other event handlers */
    return FALSE;
}

void gutter_init(GtkTextView *view, int columns){
    gutter_data_t *gutter_data;

    assert(columns >= 0);
    
    gutter_data = (gutter_data_t*)malloc(sizeof(gutter_data_t));
    gutter_data->columns = columns;
    gutter_data->gutter_pixbufs = NULL;

    /* attach our gutter data to the textview, with destructor func */
    g_object_set_data_full(G_OBJECT(view), "gutter-data", gutter_data, gutter_data_free);

    /* create the little window thing */
    gtk_text_view_set_border_window_size(view, GTK_TEXT_WINDOW_LEFT, GUTTER_COL_WIDTH * columns);
    
    /* add our redraw handler */
    g_signal_connect(G_OBJECT(view), "expose-event", G_CALLBACK(gutter_redraw_event), NULL);

    /* also handle clicks on the gutter */
    g_signal_connect(G_OBJECT(view), "button-press-event", G_CALLBACK(gutter_button_event), NULL);
}

/* draws a pixbuf into the specified "column" of the gutter
 * it gets placed next to the specified line */
/* this just adds it to the linked list and queues a redraw */
void gutter_draw_pixbuf(GtkTextView *view, int column, int line, GdkPixbuf *pbuf){
    gutter_data_t *gutter_data;
    gutter_pixbuf_t *pixbuf_data;
    gutter_pixbuf_t query;
    GSList *match;
    
    gutter_data = (gutter_data_t *)g_object_get_data(G_OBJECT(view), "gutter-data");
    assert(gutter_data != NULL);

    query.column = column;
    query.line = line;
    match = g_slist_find_custom(gutter_data->gutter_pixbufs, &query, gutter_pixbuf_cmp);

    /* only draw the thing if there isn't already a picture at that location */
    if(match == NULL){
        g_object_ref(pbuf);

        pixbuf_data = (gutter_pixbuf_t *)malloc(sizeof(gutter_pixbuf_t));
        pixbuf_data->column = column;
        pixbuf_data->line = line;
        pixbuf_data->pbuf = pbuf;

        gutter_data->gutter_pixbufs = g_slist_prepend(gutter_data->gutter_pixbufs, pixbuf_data);
        gtk_widget_queue_draw (GTK_WIDGET (view));
    }
}

void gutter_draw_stock(GtkTextView *view, int column, int line, gchar *stockid){
    GtkWidget *invis;
    GdkPixbuf *pbuf;

    invis = gtk_invisible_new();
    pbuf = gtk_widget_render_icon(invis, stockid, GTK_ICON_SIZE_MENU, "gutter");

    gutter_draw_pixbuf(view, column, line, pbuf);

    gtk_widget_destroy(invis);
}

/* clear a previously drawn pixbuf from specified location */
/* returns TRUE if it deleted anything */
gboolean gutter_clear(GtkTextView *view, int column, int line){
    gutter_pixbuf_t query;
    GSList *match;
    gutter_data_t *gutter_data;
    
    gutter_data = (gutter_data_t *)g_object_get_data(G_OBJECT(view), "gutter-data");
    assert(gutter_data != NULL);
    
    query.column = column;
    query.line = line;
    match = g_slist_find_custom(gutter_data->gutter_pixbufs, &query, gutter_pixbuf_cmp);

    if(match != NULL){
        /* remove the nasty fellow */
        g_object_unref(((gutter_pixbuf_t *)match->data)->pbuf);
        free(match->data);
        gutter_data->gutter_pixbufs = g_slist_delete_link(gutter_data->gutter_pixbufs, match);
        gtk_widget_queue_draw (GTK_WIDGET (view));
        return TRUE;
    }

    return FALSE;
}

/* clear a previously drawn pixbuf column */
/* returns how many items were deleted */
int gutter_clear_column(GtkTextView *view, int column){
    gutter_pixbuf_t query;
    GSList *match;
    gutter_data_t *gutter_data;
    int killed = 0;
    
    gutter_data = (gutter_data_t *)g_object_get_data(G_OBJECT(view), "gutter-data");
    assert(gutter_data != NULL);
    
    query.column = column;

    while((match = g_slist_find_custom(gutter_data->gutter_pixbufs, &query, gutter_pixbuf_col_cmp)) != NULL){
        /* remove the nasty fellow */
        free(match->data);
        gutter_data->gutter_pixbufs = g_slist_delete_link(gutter_data->gutter_pixbufs, match);
        killed++;
    }

    if(killed > 0){
        gtk_widget_queue_draw (GTK_WIDGET (view));
    }

    return killed;
}

