/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <libelf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <glib.h>

#include "atemu.h"
#include "breakpoints.h"
#include "../os.h"
#include "table.h"
#include "xatdb_elf.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "func_data.h"
#include "type_data.h"
#include "stabs.h"
#include "interface.h"
#include "gtk_helper.h"
#include "debug.h"
#include "support.h"

#define FNAME_STACK_SIZE 256

struct nlist {
    unsigned long n_strx;         /* index into string table of name */
    unsigned char n_type;         /* type of symbol */
    unsigned char n_other;        /* misc info (usually empty) */
    unsigned short n_desc;        /* description field */
    unsigned long n_value;        /* often an address */
};

static struct type_data *create_new_type( int chipid, char *name, char *id, char **start );

/* byte-swaps a long */
static uint32_t swap_l(uint32_t x){
    return (((x & 0x000000ff) << 24) |
            ((x & 0x0000ff00) << 8) |
            ((x & 0x00ff0000) >> 8) |
            ((x & 0xff000000) >> 24));
}

/* byte-swaps a short */
static uint16_t swap_s(uint16_t x){
    return (((x & 0x00ff) << 8) |
            ((x & 0xff00) >> 8));
}

/* converts an nlist entry of byte order order to local machine endianess */
static void convert_nlist(struct nlist *nl, unsigned order){
    if(ELF_BYTEORDER != order){
        /* only need to byteswap if byteorder differs */
        int i;
        
        nl[0].n_strx = swap_l(nl[0].n_strx);
        nl[0].n_desc = swap_s(nl[0].n_desc);
        nl[0].n_value = swap_l(nl[0].n_value);

        for(i=1; i<=nl[0].n_desc; i++){
            nl[i].n_strx = swap_l(nl[i].n_strx);
            nl[i].n_desc = swap_s(nl[i].n_desc);
            nl[i].n_value = swap_l(nl[i].n_value);
        }
    }
}

/* given the range [start, end), return a copy of the string in that range */
static char *substrdup( char *start, char *end )
{
    char *ret = NULL;

    ret = calloc( (end - start) + 1, 1 );
    if ( ret != NULL )
    {
        memcpy( ret, start, (end - start) );
    }

    return ret;
}

static void fixup_xref_helper( void *a, void *data )
{
    struct type_data * td = (struct type_data *) a;
    struct type_data * xref_td = (struct type_data *) data;

    if ( ( td->root ) && ( td->root->type == S_REFERENCE ) && 
         ( td->type != S_REFERENCE ) && ( td->root == xref_td ) )
    {
        td->root = xref_td->root;
        td->reference = xref_td->root;
    }
}

static void fixup_xrefs( int chipid, char *xref_id, struct type_data *real_ref_td )
{
    struct type_data * xref_td = NULL;

    xref_td = id_to_type_data( chipid, xref_id );

    if ( xref_td != NULL )
    {
        xref_td->root = real_ref_td;
        table_foreach( chip_array[chipid].types , fixup_xref_helper,
            xref_td );
    }
}

/* returns an allocated type id string and sets parameter 'start' to point to after id */
/* should be called before create_new_type */
static char *parse_type_id( char **start )
{
    /* curr points to the current location in the stabs string while parsing */
    char *curr = *start;
    /* temporary pointer... move along, nothing to see here */
    char *parse_temp = NULL;

//    char symbol_desc = '\0';
    char *new_type_id = NULL;

#if 0
    /* ---- parse symbol descriptor and sanity check */
    symbol_desc = (*start)[0];

    /* sanity: is a type desc. and not a local? */
    if ( ( symbol_desc == '(' ) || ( isdigit( symbol_desc ) ) )
    {
        verb_printf( TRUE, stdout, 
            _("Local symbol incorrectly classified as type desc.\n") );

        return NULL;
    }

    /* not a type('t') or structure, enumeration or union('T') description? */
    if ( ( toupper(symbol_desc) != 'T' ) )
    {
        verb_printf( TRUE, stdout, 
            _("Other than 't' or 'T' incorrectly classified as type desc.\n") );

        return NULL;
    }

    /* move past symbol descriptor */
    curr = (*start) + 1;
#endif

    /* ---- parse type info */
    
    if ( *curr == '(' )
    {
        /* gcc id '(num, num)' */

        parse_temp = strchr( curr, ')' );

        if ( parse_temp != NULL ) 
            parse_temp++;
    } else {
        /* weird id 'num' */

        strtol( curr, &parse_temp, 0 );
    }

    if ( parse_temp == NULL )
    {
        verb_printf( TRUE, stdout,  
            _("Confused by the type id [%s].\n"), curr );

        return NULL;
    }

    /* create a string for the new type id */
    new_type_id = substrdup( curr, parse_temp );

    /* moving along past '=' */
    (*start) = parse_temp;

    return new_type_id;
}

static struct type_data *parse_type( int chipid, char **start )
{
    struct type_data * td = NULL;
    char * type_id = NULL;

    type_id = parse_type_id( start );

    if ( type_id == NULL )
    {
        verb_printf( TRUE, stdout,
            _("Malformed type definition (no '=').\n") );

        return NULL;
    }

    if ( **start == '=' )
    {
        /* ---- nested type definition */
        td = create_new_type( chipid, "", type_id, start );
    } else {
        /* ---- 'simple' reference */
        td = id_to_type_data( chipid, type_id );
    }

    free(type_id);

    return td;
}

/* returns a pointer to the ';' at the end of an AIX extension */
static char *parse_aix_extensions( struct type_data *td, char *aix_start )
{
    int align;
    int aix_size;
    char *curr = aix_start;

    switch( curr[1] )
    {
        /* alignment? */
        case 'a':
            align = atoi( &curr[2] );
            verb_printf( FALSE, stdout, _("\tAlignement: %d\n"),
                align );
            break;

        /* pointer? */
        case 'p':
            verb_printf( FALSE, stdout, _("\tPointer.\n") );
            break;

        /* packed? */
        case 'P':
            verb_printf( FALSE, stdout, _("\tPacked.\n") );
            break;

        /* pre-sized? */
        case 's':
            aix_size = atoi( &curr[2] );
            verb_printf( FALSE, stdout, _("\tSize: %d\n"), aix_size );
            break;

        /* string? */
        case 'S':
            verb_printf( FALSE, stdout, _("\tString.\n") );
            break;

        /* unknown? */
        default:
            verb_printf( FALSE, stdout, _("\tUnknown AIX.\n") );
            break;
    }

    curr = strchr( curr, ';' );

    return curr;
}

/* helper for create_new_type */
static struct type_data *create_new_reference( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    struct type_data * ref_td = NULL;

    ref_td = parse_type( chipid, start );

    /* create a new type (finally) */
    td = create_type_data( name, id, 0, 0, 0, REFERENCE );

    /* here, link to your references type, that way chained refernces are
       squashed and we don't have to go following silly linked list type
       things... */
    if ( ref_td )
        td->root = ref_td->root;
    else
        td->root = NULL;

    /* this DOES create a linked list, but just for naming. */
    td->reference = ref_td;

    /* insert in table */
    add_type_data( chipid, td );

    if ( ref_td )
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Reference -> (%s)<%s>\n",
             (name)?(name):("(null)"), 
            id, (ref_td->name)?(ref_td->name):("(null)"), ref_td->id );
    else
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Reference -> (%s)<%s>\n",
            (name)?(name):("(null)"), 
            id, "null", "null" );

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_range( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;

    struct type_data * range_td = NULL;
//    char * range_id = NULL;

    /* For determining byte size of ranges. */
    long long int begin, end;
    long long unsigned int range;
    int bits, bytes;

    range_td = parse_type( chipid, start );

    if ( **start != ';' )
    {
        verb_printf( TRUE, stdout, _("Malformed range type definition (missing first ';').\n") );

        return NULL;
    }
    (*start)++;

    begin = strtoll( *start, start, 0 );

    if ( **start != ';' )
    {
        verb_printf( TRUE, stdout, _("Malformed range type definition (missing second ';').\n") );

        return NULL;
    }
    (*start)++;

    end = strtoll( *start, start, 0 );

    if (( begin > 0 ) && ( end == 0 ))
    {
        bytes = begin;
        bits = bytes * 8;
    } else {
        if (( begin > 0 ) && ( end > 0 ))
        {
            begin = -begin;
        }

        range = end - begin;

        /* A poor man's lg function... */
        bits = 0;
        while( range )
        {
            bits++;
            range >>= 1;
        }

        bytes = (int) ceil( (double) bits / 8 );
    }

    /* we don't know the type, be we're going to guess it is signed and
       the create_type_data function will fix it up. */
    td = create_type_data( name, id, bytes, begin, end, SIGNED_NUM );

    if ( td == NULL )
    {
        verb_printf( TRUE, stdout, _("Couldn't create type_data.\n") );

        return NULL;
    }

    /* if this doesn't have a type id (td->id == ""), don't save in table */
    if ( strlen( td->id ) != 0 )
    {
        add_type_data( chipid, td );
    }

    if ( range_td )
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Range -> (%s)<%s> |%d|\n", (name)?(name):("(null)"), id,
            range_td->name, range_td->id, bytes );
    else
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Range -> (%s)<%s> |%d|\n", (name)?(name):("(null)"), id,
            "null", "null", bytes );

    /* move past ';' */
    (*start)++;

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_array( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    struct type_data * index_td = NULL;
    struct type_data * element_td = NULL;
 
    /* size of array in bytes */
    uint16_t size = 0;
    long long int begin = 0, end = 0;

    /* arrays are, uh, "underspecified", so we assume (as GDB does) 
       that the index is a range type */
    if ( **start != TD_RANGE )
    {
        verb_printf( TRUE, stdout, _("Array indexed by non-range type (unsupported).\n") );

        return NULL;
    }
    (*start)++;

    index_td = create_new_range( chipid, "", "", start );

    element_td = parse_type( chipid, start );

    if (( element_td != NULL ) && ( index_td != NULL ))
    {
        if ( ( element_td->root != NULL )
             && ( index_td->root != NULL ) )
        {
            begin = index_td->root->begin;
            end = index_td->root->end;

            size = ( end - begin + 1 );
            size *=  element_td->root->size;
        }
    }

    td = create_type_data( name, id, size, begin, end, ARRAY );

    if ( td == NULL )
    {
        verb_printf( TRUE, stdout, _("Couldn't create type_data.\n") );

        return NULL;
    }

    td->extra = element_td;

    add_type_data( chipid, td );

    verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Array [%lld - %lld] |%d| \n",
        (name)?(name):("(null)"), id,
        begin, end, size );

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_struct( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    uint16_t size = 0;

    size = strtoll( *start, start, 0 );

    td = create_type_data( name, id, size, 0, 0, STRUCT );

    if ( td == NULL )
    {
        verb_printf( TRUE, stdout, _("Couldn't create type_data.\n") ); 

        return NULL;
    }

    fixup_xrefs( chipid, name, td );

    add_type_data( chipid, td );

    verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Structure |%d|\n", 
        name, id, size );

    while( **start != ';' )
    {
        struct type_data * field_td = NULL;
        struct field * f = NULL;
        char *field_name, *field_name_end;
        uint16_t bit_length = 0, bit_start = 0;

        field_name_end = strchr( *start, ':' );
        if ( field_name_end == NULL )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed struct (missing ':').\n") );

            free_type_data( td );

            return NULL;
        }

        field_name = substrdup( *start, field_name_end );

        /* move past ':' */
        (*start) = field_name_end + 1;

        field_td = parse_type( chipid, start );

        if ( **start != ',' )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed struct (missing first ',').\n") );

            free_type_data( td );
            free( field_name );

            return NULL;
        }
        /* move past ',' */
        (*start)++;

        bit_start = strtoll( *start, start, 0 );

        if ( **start != ',' )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed struct (missing second ',').\n") );

            free_type_data( td );
            free( field_name );

            return NULL;
        }
        /* move past ',' */
        (*start)++;

        bit_length = strtoll( *start, start, 0 );

        if ( **start != ';' )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed struct (missing end of field ';').\n") );

            free_type_data( td );
            free( field_name );

            return NULL;
        }
        /* move past ';' */
        (*start)++;

        f = create_field( field_name, field_td, bit_start, bit_length, 0 );

        if ( field_td )
            verb_printf( FALSE, stdout, "\t[Created field] (%s)<(%s)<%s>> [%d] |%d|\n",
                field_name, field_td->name, field_td->id, bit_start, 
                bit_length );
        else
            verb_printf( FALSE, stdout, "\t[Created field] (%s)<(%s)<%s>> [%d] |%d|\n",
                field_name, "(null)", "(null)", bit_start, bit_length );

        free( field_name );

        add_field_to_type( td, f );
    }

    /* move past end ';' */
    (*start)++;

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_enum( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;

    /* according to the usual source, stabs info pages, enums are usually 32 bits... */
    td = create_type_data( name, id, 4, 0, 0, ENUM );
    if ( td == NULL )
    {
        verb_printf( TRUE, stdout, _("Couldn't create type_data.\n") ); 

        return NULL;
    }

    verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Enum\n", name, id);

    while( **start != ';' )
    {
        struct field * f = NULL;
        char *field_name, *field_name_end;
        uint16_t value = 0;

        field_name_end = strchr( *start, ':' );
        if ( field_name_end == NULL )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed enum (missing ':').\n") );

            free_type_data( td );

            return NULL;
        }

        field_name = substrdup( *start, field_name_end );

        /* move past ':' */
        (*start) = field_name_end + 1;

        value = strtoll( *start, start, 0 );

        if ( **start != ',' )
        {
            verb_printf( TRUE, stdout, 
                _("Malformed enum (missing ',').\n") );

            free_type_data( td );
            free( field_name );

            return NULL;
        }
        /* move past ',' */
        (*start)++;

        f = create_field( field_name, NULL, 0, 0, value );

        verb_printf( FALSE, stdout, "\t[Created field] (%s) = [%d]\n", 
            field_name, value );

        free( field_name );

        add_field_to_type( td, f );
    }

    add_type_data( chipid, td );

    /* move past end ';' */
    (*start)++;

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_pointer( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    struct type_data * pointer_td = NULL;

    pointer_td = parse_type( chipid, start );

    /* create a new type (finally) */
    td = create_type_data( name, id, 2, 0, 0, POINTER );

    td->extra = pointer_td;

    if ( pointer_td )
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Pointer -> (%s)<%s>\n",
            name, id, pointer_td->name, pointer_td->id );
    else
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Pointer -> (%s)<%s>\n",
            name, id, "null", "null" );

    add_type_data( chipid, td );

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_function( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    struct type_data * pointer_td = NULL;

    pointer_td = parse_type( chipid, start );

    /* create a new type (finally) */
    td = create_type_data( name, id, 2, 0, 0, FUNCTION );

    td->extra = pointer_td;

    if ( pointer_td )
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Function -> returns (%s)<%s>\n",
            name, id, pointer_td->name, pointer_td->id );
    else
        verb_printf( FALSE, stdout, "[Created type] (%s)<%s> Function -> returns (%s)<%s>\n",
            name, id, "null", "null" );

    add_type_data( chipid, td );

    return td;
}

/* helper for create_new_type */
static struct type_data *create_new_xref( int chipid, char *name, char *id, char **start )
{
    struct type_data * td = NULL;
    char * xref_name_end = NULL;
    char * xref_name = NULL;

    /* go past s or E or u */
    (*start)++;

    /* find end of name */
    xref_name_end = strchr( *start, ':' );
    if ( xref_name_end == NULL )
    {
        verb_printf( TRUE, stdout, 
            _("Malformed cross reference (missing ':').\n") );

        return NULL;
    }

    xref_name = substrdup( *start, xref_name_end );

    /* create a new type (finally) */
    td = create_type_data( name, id, 0, 0, 0, REFERENCE );
    td->root = create_type_data( "", xref_name, 0, 0, 0, S_REFERENCE );
    td->reference = td->root;

    verb_printf( FALSE, stdout, "[Created type] Cross Reference to <%s>\n", xref_name );

    free( xref_name );

    /* move past ':' */
    *start = xref_name_end + 1;

    add_type_data( chipid, td );
    add_type_data( chipid, td->root );

    return td;
}

/*
 * create_new_type:
 *
 *     Turns a stabs type definition of the form:
 *       "=s4real:(0,1),0,16;imag:(0,1),16,16;;"
 *     into a meaningful structure. Also inserts into the correct table.
 *
 */
static struct type_data *create_new_type( int chipid, char *name, char *id, char **start )
{
    /* td is a pointer to the type data that will be created (or looked up)
       and returned. */
    struct type_data *td = NULL;

    /* curr points to the current location in the stabs string while parsing */
    char *curr = NULL;

    char type_desc = '\0';

    /* ---- sanity check on input */
    if ( **start != '=' )
    {
        verb_printf( TRUE, stdout, 
            _("'start' param. bad (no '=' as first char.) entering create_new_type.\n") );

        return NULL;
    }

    /* moving along past '=' */
    curr = *start + 1;

    /* ---- parse type description */
    type_desc = *curr;

    /* AIX type attribute extensions? */
    if ( type_desc == TD_TYPE_ATTR )
    {
        curr = parse_aix_extensions( td, curr );

        if ( curr == NULL )
        {
            verb_printf( TRUE, stdout, 
                _("AIX extension incorrectly formed, no ';'.\n") );

            return NULL;
        }

        /* move curr to after AIX ';' */
        curr++;

        type_desc = *curr;
    }

    /* a reference or a type description */
    if ( ( type_desc == '(' ) || ( isdigit( (int)type_desc ) ) )
    {
        /* ---- type reference */
        td = create_new_reference( chipid, name, id, &curr );
    } else {

        /* move curr past type descriptor */
        curr++;

        switch( type_desc )
        {
            case TD_RANGE:
                td = create_new_range( chipid, name, id, &curr );
                break;

            case TD_ARRAY:
                td = create_new_array( chipid, name, id, &curr );
                break;

            case TD_UNION:
            case TD_STRUCT:
                td = create_new_struct( chipid, name, id, &curr );
                break;

            case TD_ENUM:
                td = create_new_enum( chipid, name, id, &curr );
                break;

            case TD_POINTER:
                td = create_new_pointer( chipid, name, id, &curr );
                break;

            case TD_FUNCTION:
                td = create_new_function( chipid, name, id, &curr );
                break;

            /* stuff that we don't care about -> turns into a simple reference */
            case TD_CONST:
            case TD_VOLATILE_QUAL:
                td = create_new_reference( chipid, name, id, &curr );
                break;

            case TD_CROSS_REF:
                td = create_new_xref( chipid, name, id, &curr );
                break;

            default:
                verb_printf( TRUE, stdout,
                    _("\tUnimplemented type ['%c'].\n"), type_desc );

                /* cause style is... cool? */
                break;
        }
    }

    *start = curr;

    return td;
}

static void create_type( int chipid, char *name, char *types )
{
    char *new_type;
    char *start = types + 1;
/*
    char new_desc;
    char *curr, *semi, *comma;
    char type_desc;
    char *type_def;

    struct type_data *td;
    struct type_data *ref;
    struct field *f;
*/
    /* For determining byte size of ranges. */
/*
    long long int begin, end;
    long long unsigned int range;
    int bits, bytes, size = 0, align = 0;
*/

    /* TODO: remember to get rid of the '+ 1' for start and actually 
       check for 't'/'T' */
    new_type = parse_type_id( &start );
    create_new_type( chipid, name, new_type, &start );
    free(new_type);
}

/* this parses the stab list you pass it, storing things in nice structs */
static void build_maps( char * strings, struct nlist * nl, int chipid ){
    /* there are n_desc MORE stab entries after this one */
    int maxstabs=nl[0].n_desc;
    int i;
    unsigned long func_addr = 0;
    unsigned char in_func = 0;
    unsigned char in_c_line = 0;
    unsigned long old_addr = 0;
    unsigned long curr_line = 0;
    struct func_data *fd = NULL;
    int len;
    int curr_id = 0;

    int is_param;
    int type;

    char *fname_path = "";
    char *fname_stack[FNAME_STACK_SIZE];
    int fname_index=0;
    
    for(i=0; i < FNAME_STACK_SIZE; i++){
        fname_stack[i] = NULL;
    }

    for(i=1; i <= maxstabs; i++){
        unsigned long off = nl[i].n_strx;
        char *stab, *data;

        stab = strdup(strings + off);
        
        switch(nl[i].n_type){
            case N_FUN:
                /* Function */
                if ( !in_func ) {
                    /* start of a function */
                    func_addr = nl[i].n_value;

                    if ( ( data = strchr( stab, ':' ) ) )
                        *(data++) = 0;

                    fd = create_func_data( stab, func_addr, func_addr );

                    in_func = 1;
                } else {
                    /* end of last function started */

                    /* end any line mapping in progress */
                    if ( in_c_line ) {
                        in_c_line = 0;

                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value + func_addr, curr_line, fname_stack[fname_index] );
                    }

                    if ( fd != NULL )
                    {
                        fd->addr_end = func_addr + nl[i].n_value;

                        add_func_data( chipid, fd );
                    }

                    in_func = 0;
                    curr_id = 0;
                }
                break;

            case N_STSYM:
            case N_GSYM:
                {
                    char *curr = stab;
                    char *global_name = NULL, *global_name_end = NULL;

                    verb_printf( FALSE, stdout, _("Global: [%s]\n"), stab );

                    if ( ( global_name_end = strchr( curr, ':' ) ) != NULL )
                    {
                        global_name = substrdup( curr, global_name_end );
                    } else { 
                        verb_printf( TRUE, stdout, _("Invalid GSYM or STSYM stabs string field, missing ':'.\n") );
                        break;
                    }

                    curr = global_name_end + 1;

                    /* make sure it's a global or static symbol */
                    if (( *curr == SD_GLOBAL_VAR ) || ( *curr == SD_FILE_SCOPE_VAR))
                    {
                        struct type_data *td;

                        /* move past symbol desc. */
                        curr++;

                        td = parse_type( chipid, &curr );

                        size_global( chipid, global_name, td );
                    }

                    free( global_name );
                }
                break;

            case N_RSYM:
            case N_LSYM:
            case N_PSYM:
                /* Locals and Params */

                /* TODO: rewrite this sewage... */

                if ( ( data = strchr( stab, ':' ) ) )
                    *(data++) = 0;

                if ( in_func ) 
                {
                    struct type_data *td;

                    if ( fd != NULL )
                    {
                        if ( nl[i].n_type == 0x40 ) {
                            if ( *data == 'P' ) {
                                is_param = 1;
                            } else if ( *data == 'r' ) {
                                is_param = 0;
                            }

                            data++;

                            type = LOCAL_TYPE_REG;
                        } else if ( nl[i].n_type == 0xa0 ) {
                            if ( *data == 'p' ) {
                                is_param = 1;
                            } else {
                                is_param = 0;
                            }

                            data++;

                            type = LOCAL_TYPE_STACK;
                        } else {
                            is_param = 0;

                            type = LOCAL_TYPE_STACK;
                        }

                        td = parse_type( chipid, &data );

                        create_and_add_local( fd, stab, is_param, type,
                            nl[i].n_value, td, curr_id++, NULL );
                    }
                } else if ( nl[i].n_type == N_LSYM ) {
                    if ( *data != '(' )
                    {
                        create_type( chipid, stab, data );
                    }
                }
                break;
            case N_EINCL:
                /* end include file */
                /* this one is stack-based */

                free(fname_stack[fname_index]);
                
                fname_index--;
                if(fname_index < 0){
                    verb_printf( TRUE, stdout, _("decremented stack below 0, oh no!\n") );
                    fname_index = 0;
                }
                break;
            case N_BINCL:
                /* begin include file */
                /* this one is stack-based */

                fname_index++;
                if(fname_index >= FNAME_STACK_SIZE){
                    verb_printf( TRUE, stdout, _("incremented stack above %d, oh no!\n"), FNAME_STACK_SIZE);
                    fname_index = FNAME_STACK_SIZE - 1;
                }
                /* absolute paths start with either "/" or "X:" */
                if((strings + off)[0] != '/' && !( isalpha((int)(strings + off)[0]) && (strings + off)[1] == ':')){
                    fname_stack[fname_index] = malloc(strlen(fname_path) + strlen(strings + off) + 1);
                    // non 'n' versions ok, as lengths carefully calculated
                    strcpy(fname_stack[fname_index], fname_path);
                    strcat(fname_stack[fname_index], strings + off);
                }else{
                    fname_stack[fname_index] = strdup(strings + off);
                }
                add_cfile(chipid, fname_stack[fname_index]);
                break;        
            case N_SOL:
            case N_SO:
                /* start of C/include file */

                /* end any line mapping in progress */
                if ( in_c_line ) {
                    in_c_line = 0;

                    add_addr_file_mapping( chipid, old_addr,
                            nl[i].n_value, curr_line, fname_stack[fname_index] );
                }

                len = strlen(strings + off);
                /* last character of (strings + off) is a '/', this is a path */
                if( len > 0 && (strings + off)[len - 1] == '/'){
                    fname_path = strings + off;
                }else{
                    /* it's the actual file name */
                    free(fname_stack[fname_index]);
                    /* absolute paths start with either "/" or "X:" */
                    if(len > 0 && (strings + off)[0] != '/' && !( len > 1 && isalpha((int)(strings + off)[0]) && (strings + off)[1] == ':')){
                        fname_stack[fname_index] = malloc(strlen(fname_path) + strlen(strings + off) + 1);
                        // non 'n' versions ok, as lengths carefully calculated
                        strcpy(fname_stack[fname_index], fname_path);
                        strcat(fname_stack[fname_index], strings + off);
                    }else{
                        fname_stack[fname_index] = strdup(strings + off);
                    }
                    if(len > 0){
                        add_cfile(chipid, fname_stack[fname_index]);
                    }
                }
                break;
            case N_SLINE:
                /* Line Number */
                if ( in_c_line ) {
                    if ( in_func ) {
                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value + func_addr, curr_line, fname_stack[fname_index] );
                    } else {
                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value, curr_line, fname_stack[fname_index] );
                    }
                }

                if ( in_func ) {
                    curr_line = nl[i].n_desc;
                    old_addr = nl[i].n_value + func_addr;

                    create_and_add_c_line( fd, old_addr, curr_line );
                } else {
                    curr_line = nl[i].n_desc;
                    old_addr = nl[i].n_value;
                }

                in_c_line = 1;
                break;
            default:
//                printf("Unhandled stab: type 0x%02x\n",nl[i].n_type);
                break;
        }

        free(stab);
        // prevent the UI from becoming unresponsive
        g_main_context_iteration(NULL, FALSE);
        if(xatdb_quit){
            return;
        }
    }

    /* free any strings left in the stack */
    for(i=0; i <= fname_index; i++){
        free(fname_stack[i]);
    }

    /* make the C stepping breakpoints */
    generate_auto_breakpoints(chipid);

    fflush(stdout);
}

int parse_elf(const char * elfname, int chipid ){
    int elffile;
    Elf *elf;
    Elf32_Shdr *shdr;
    Elf32_Ehdr *ehdr;
    Elf_Scn *scn, *stab=NULL, *stabstr=NULL, 
        *symtab = NULL, *strtab = NULL;
    Elf_Data *data;
    GString *parse_message = g_string_new( "" );

    int cnt;
    struct nlist *nl;
    char *strings;

    GtkTextMark *mark;

    if(xatdb_quit){
        return -1;
    }

    elffile=open(elfname, O_RDONLY);
    if(elffile < 0){
        perror(elfname);
        verb_printf( TRUE, stdout, _("Error opening %s!\n"), elfname);
        return -1;
    }

    /* set version of elf */
    elf_version(EV_CURRENT);

    /* open the elf file */
    elf=elf_begin(elffile, ELF_C_READ, NULL);
    if(elf == NULL){
        verb_printf( TRUE, stdout, "%s: %s.\n", elfname, elf_errmsg(elf_errno()));
        close(elffile);
        return -1;
    }
    
    /* give the user feedback in the status bar */
    g_string_append( parse_message, _("Parsing symbols in ") );
    g_string_append( parse_message, elfname );
    g_string_append( parse_message, _("...") );
    
    gtk_statusbar_pop( GTK_STATUSBAR( statusbar1 ), 1 );
    gtk_statusbar_push( GTK_STATUSBAR( statusbar1 ), 1, parse_message->str );
    
    gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 
        gtk_progress_bar_get_fraction( GTK_PROGRESS_BAR( progressbar1 ) ) 
        + (double)( 0.5 / ( 0.5 + ( 2.0 * (double)chip_array_cnt ) ) ) );
    
    /* print feedback to the console */
    mark = text_buffer_printf( textbuff_console, "%s\n", parse_message->str );
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);

    g_string_free( parse_message, TRUE );

    /* get some sections */
    if (((ehdr = elf32_getehdr(elf)) == NULL) ||
        ((scn = elf_getscn(elf, ehdr->e_shstrndx)) == NULL) ||
        ((data = elf_getdata(scn, NULL)) == NULL)){
        verb_printf( TRUE, stdout, "%s: %s.\n", elfname, elf_errmsg(elf_errno()));
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* print names of sections */
    for (cnt = 1, scn = NULL; (scn = elf_nextscn(elf, scn)); cnt++) {
        if ((shdr = elf32_getshdr(scn)) == NULL){
            verb_printf( TRUE, stdout, "%s: %s.\n", elfname, elf_errmsg(elf_errno()));
            elf_end(elf);
            close(elffile);
            return -1;
        }
        if(strcmp((char *)data->d_buf + shdr->sh_name, ".stab") == 0){
            stab = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".stabstr") == 0){
            stabstr = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".strtab") == 0){
            strtab = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".symtab") == 0){
            symtab = scn;
        }
    }

    /* do the symbols */
    if ( (strtab == NULL) || (symtab == NULL) )
    {
        verb_printf( TRUE, stdout, _("Missing either .symtab or .strtab.  Ergo, no symbol info.\n") );
    } else {
        /* borrowed from: 
           http://developers.sun.com/solaris/articles/elf.html */

        Elf32_Sym *esym, *lastsym;
        char *name;

        data = elf_getdata(strtab, NULL);
        strings = (char *) data->d_buf;
        data = elf_getdata(symtab, NULL);
        shdr = elf32_getshdr(symtab);

        esym = (Elf32_Sym*) data->d_buf;
        lastsym = (Elf32_Sym*) ((char *)data->d_buf + data->d_size);

        /* now loop through the symbol table */
        for (; esym < lastsym; esym++)
        {
            if ((esym->st_value == 0) ||
                (ELF32_ST_BIND(esym->st_info)== STB_WEAK) ||
                (ELF32_ST_BIND(esym->st_info)== STB_NUM) ||
                (ELF32_ST_TYPE(esym->st_info)!= STT_OBJECT)) {
                continue;
            }

            name = elf_strptr(elf,shdr->sh_link, (size_t)esym->st_name);

            /* add to global table */
            if( name != NULL )
            {
                verb_printf( FALSE, stdout, "%s : <%08lx>\n", name, 
                    (long unsigned int) esym->st_value & (~0x800000) );

                add_global( chipid, name, esym->st_value & (~0x800000), NULL );
            }
            // prevent the UI from becoming unresponsive
            g_main_context_iteration(NULL, FALSE);
            if(xatdb_quit){
                elf_end(elf);
                close(elffile);
                return -1;
            }
        }
    }

    if(stab == NULL || stabstr == NULL){
        verb_printf( TRUE, stdout, _("Missing either .stab or .stabstr.  Ergo, no debug info.\n") );
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* get .stab data section (fixing endianness as we go) */
    data = elf_getdata(stab, NULL);
    nl=(struct nlist*)data->d_buf;
    convert_nlist(nl, ehdr->e_ident[EI_DATA]);

    /* get the stab string table data section */
    data = elf_getdata(stabstr, NULL);
    strings=(char *)data->d_buf;

    build_maps( strings, nl, chipid );
    if(xatdb_quit){
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* do the cfile sort thing */
    make_cfile_menu(chipid);

    elf_end(elf);
    close(elffile);
    return 0;
}
