/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>

/* for the CONVERT_?BYTE macros */
#include "../defs.h"

#include "type_data.h"
#include "debug.h"

struct type_data *create_type_data( char *name, char *id, uint16_t size,
    long long int begin, long long int end, enum td_type type )
{
    struct type_data *ret;

    ret = (struct type_data *) malloc( sizeof(struct type_data) );
    if ( ret == NULL )
    {
        return NULL;
    }

    /* Empty list */
    ret->fields = NULL;

    ret->name = strdup( name );
    ret->id = strdup( id );

    ret->type = type;

    ret->size = size;
    ret->begin = begin;
    ret->end = end;

    ret->root = ret;
    ret->reference = ret;

    if ( (( ret->begin != 0 ) || ( ret->end != 0 )) &&
         (( type == SIGNED_NUM ) || ( type == UNSIGNED_NUM )) )
    {
        if ( ret->begin < 0 )
            ret->type = SIGNED_NUM;
        else if ( ret->begin >= 0 )
            ret->type = UNSIGNED_NUM;
    }

    return ret;
}

void add_field_to_type( struct type_data * td, struct field * f )
{
    if (( td == NULL ) || ( f == NULL ))
    {
        verb_printf( TRUE, stdout, "Passed NULL arg to add_field_to_type.\n" );
        return;
    }

    td->fields = g_slist_append( td->fields, (gpointer) f );
}

int type_data_compare_id( const void * a, const void * b )
{
    struct type_data *atd = (struct type_data *) a;
    struct type_data *btd = (struct type_data *) b;

    return ( strcmp( atd->id, btd->id ) );
}

void free_type_data( void * a )
{
    struct type_data *td = (struct type_data *) a;
    GSList *curr;

    if ( td == NULL )
        return;

    for( curr = td->fields; curr != NULL; curr = curr->next )
        free_field( curr->data );

    g_slist_free( td->fields );

    free( td->name );
    free( td->id );
    free( td );
}

struct field * create_field( char * name, struct type_data *td,
    uint16_t bit_offset, uint16_t bit_length, uint16_t value )
{
    struct field * ret;

    ret = (struct field *) malloc( sizeof(struct field) );
    if ( ret == NULL )
    {
        return NULL;
    }

    ret->name = strdup( name );
    ret->type = td;

    ret->bit_offset = bit_offset;
    ret->bit_length = bit_length;
    ret->value = value;

    return ret;
}

void free_field( struct field * f )
{
    if ( f )
        free( f->name );

    free( f );
}

static int render_signed_int( char *buff, int size, uint8_t *data, int int_size )
{
    int to_write;
    
    switch( int_size )
    {
        case 8:
            to_write = snprintf( buff, size, "%ld", 
                (int64_t)CONVERT_8BYTE(*((uint64_t*) data) ) );
            break;

        case 4:
            to_write = snprintf( buff, size, "%d", 
                (int32_t)CONVERT_4BYTE(*((uint32_t*) data) ) );
            break;

        case 2:
            to_write = snprintf( buff, size, "%d",
                (int16_t)CONVERT_2BYTE(*((uint16_t*) data) ) );
            break;

        case 1:
            to_write = snprintf( buff, size, "%d", *((int8_t *) data) );
            break;
	
        default:
            to_write = snprintf( buff, size, "Error: Number weird size = %d.",
               int_size );
            break;
    }

    return ( size - to_write );
}

static int render_unsigned_int( char *buff, int size, uint8_t *data, int int_size )
{
    int to_write;

    switch( int_size )
    {
        case 8:
            to_write = snprintf( buff, size, "%ld", 
                CONVERT_8BYTE(*((uint64_t*) data) ) );
            break;

        case 4:
            to_write = snprintf( buff, size, "%d", 
                CONVERT_4BYTE(*((uint32_t*) data) ) );
            break;

        case 2:
            to_write = snprintf( buff, size, "%d",
                CONVERT_2BYTE(*((uint16_t*) data) ) );
            break;

        case 1:
            to_write = snprintf( buff, size, "%d", *((uint8_t *) data) );
            break;
	
        default:
            to_write = snprintf( buff, size, "Error: Number weird size = %d.",
               int_size );
            break;
    }

    return ( size - to_write );
}

static int render_hexdump( char *buff, int size, uint8_t *data, int count )
{
    int i, to_write;

    for( i = 0; i < count; i++ )
    {
        if ( i != (count - 1) )
          to_write = snprintf( buff, size, "%02x ", data[i] );
        else
          to_write = snprintf( buff, size, "%02x", data[i] );

        size -= to_write;

        if (size < 0)
            return size;

        buff += to_write;
    }

    return size;
}

static int render_type_name( char *buff, int size, struct type_data *td )
{
    int to_write;
    int ret;

    if ( td != NULL )
    {
        switch( td->type )
        {
            case POINTER:
                {
                    char pbuff[256];

                    /* TODO: check the return */
                    render_type_name( pbuff, 256, td->extra );

                    ret = snprintf( buff, size, "(%s *)", pbuff );
                    ret = size - ret;
                }
                break;

            case FUNCTION:
                {
                    char pbuff[256];

                    render_type_name( pbuff, 256, td->extra );

                    ret = snprintf( buff, size, "(%s)()", pbuff );
                    ret = size - ret;
                }
                break;

            case REFERENCE:
                if ( ( td->name == NULL ) || ( td->name[0] == '\0') )
                {
                    ret = render_type_name( buff, size, td->reference );
                } else {
                    ret = snprintf( buff, size, "%s", td->name );
                    ret = size - ret;
                }
                break;

            case STRUCT:
                if ( ( td->name == NULL ) || ( td->name[0] == '\0') )
                {
                    ret = snprintf( buff, size, "struct (no name)" );
                } else {
                    ret = snprintf( buff, size, "struct %s", td->name );
                }
                ret = size - ret;
                break;

            case ARRAY:
                {
                    char pbuff[256];

                    if ( td->extra == NULL )
                    {
                        ret = snprintf( buff, size, "(no name)[%lld]", td->end + 1 );
                    } else {
                        /* TODO: check the return */
                        render_type_name( pbuff, 256, td->extra );
                        ret = snprintf( buff, size, "%s[%lld]", pbuff, td->end + 1);
                    }

                    ret = size - ret;
                }
                break;

            default:
                ret = snprintf( buff, size, "%s", td->name );
                ret = size - ret;
                break;
        }
    } else {
        ret = snprintf( buff, size, "void" );
        ret = size - ret;
    }

    return ret;
}

static int render_pointer( char *buff, int size, uint8_t *data, struct type_data *td )
{
    int to_write;
    char p_void[] = "void";
    char *p_type;
    char p_name[256];

    if ( ( td != NULL ) && ( td->extra != NULL ) )
        p_type = td->extra->name;
    else
        p_type = p_void;

    render_type_name( p_name, 256, td );
    to_write =  snprintf( buff, size, "%s 0x%08x", p_name, 
        CONVERT_2BYTE( *((uint16_t *) data) ) );

    return ( size - to_write );
}

static int render_enum( char *buff, int size, uint8_t *data, struct type_data *td )
{
    int to_write;
    char e_invalid[] = "Invalid ENUM";
    char *e_name;
    uint16_t value;

    value = CONVERT_2BYTE( *((uint16_t *) data) );

    e_name = e_invalid;

    if ( ( td != NULL ) && ( td->fields != NULL ) )
    {
        GSList *curr = td->fields;
        struct field *f = NULL;

        while( curr != NULL )
        {
            f = (struct field *) curr->data;

            if ( f->value == value )
            {
                e_name = f->name;

                break;
            }

            curr = curr->next;
        }
    }

    to_write =  snprintf( buff, size, "%s = %d", e_name, value );

    return ( size - to_write );
}

int render_type( char *buff, int size, uint8_t *data, struct type_data *td )
{
    int ret = -1;

    if ( data == NULL )
    {
        ret = snprintf( buff, size, "Not enough information." );
        ret = size - ret;
    } else if ( td == NULL ) {
        ret = render_hexdump( buff, size, data, 4 );
    } else {
        switch( td->type )
        {
            case SIGNED_NUM:
                ret = render_signed_int( buff, size, data, td->size );
                break;

            case UNSIGNED_NUM:
                ret = render_unsigned_int( buff, size, data, td->size );
                break;

            case POINTER:
                ret = render_pointer( buff, size, data, td );
                break;

            case REFERENCE:
                if (( td->root->type == SIGNED_NUM ) || ( td->root->type == UNSIGNED_NUM ))
                {
                    ret = render_type( buff, size, data, td->root );
                } else {
                    ret = render_type_name( buff, size, td );
                }
                break;

            case ENUM:
                ret = render_enum( buff, size, data, td );
                break;

            case STRUCT:
                ret = render_type_name( buff, size, td );
                break;

            case ARRAY:
                ret = render_type_name( buff, size, td );
                break;

            case UNKNOWN:
                ret = snprintf( buff, size, "Unknown type." );
                ret = size - ret;
                break;

            default:
                ret = render_hexdump( buff, size, data, 4 );
                break;
        }
    }

    return ret;
}
