/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "watchpoints.h"
#include "atemu.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "interface.h"
#include "debug.h"

/* set the visibility of the watchpoint icon on a global */
void watch_set_icon( gboolean sense, int chipid, GtkTreeIter iter )
{
    /*  sense       whether or not an item is being added or removed
     *  chipid      the chip to act upon
     *  iter        a TreeIter to the global being acted upon */

    if( sense )
    {
        gtk_tree_store_set( chip_array[curr_chip].globals, &iter, GLOBALS_ICON, XATDB_STOCK_WATCH, -1);
    }
    else
    {
        gtk_tree_store_set( chip_array[curr_chip].globals, &iter, GLOBALS_ICON, "", -1);
    }
}

/* set the sensitivity of watchpoint related menu items as appropriate */
void watch_set_sensitive( gboolean sense, int chipid )
{
    /*  sense       whether or not an item is being added or removed
     *  chipid      the chip to act upon */
    
    if( curr_chip == chipid )
    {
        if( GTK_WIDGET_HAS_FOCUS( chip_array[curr_chip].treeview_globals ) )
        {
            gtk_widget_set_sensitive( menu_debug_clear_watch, sense );
            gtk_widget_set_sensitive( menu_debug_set_watch, !sense );
        }
        
        if( chip_array[curr_chip].num_watches <= 0 )
        {
            gtk_widget_set_sensitive( menu_debug_clear_watches, FALSE );
            gtk_widget_set_sensitive( menu_debug_clear_watch, FALSE );
        }
        else
        {
            gtk_widget_set_sensitive( menu_debug_clear_watches, TRUE );
        }
    }
}

/* add a watchpoint */
/* possibly highlight the global variable being watched */
void add_watchpoint(int chipid, long addr, struct tuple_global_addr *glob )
{
    char hex[32];
    char *glob_name = NULL;
    int glob_size = XATDB_DEFAULT_VAR_SIZE;
    int glob_id;
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    gtk_list_store_append( chip_array[chipid].watchp, &iter );

    if(glob != NULL){
        glob_name = glob->name;
        glob_id = glob->notify_id;

        if ( glob->td != NULL )
        {
            if ( glob->td->root != NULL )
            {
                glob_size = glob->td->root->size;
            } else {
                /* CHECK: shouldn't... log? */
                glob_size = glob->td->size;
            }
        }
    }

    snprintf( hex, 32, "0x%08lx", addr );
    hex[31] = '\0';

    gtk_list_store_set( GTK_LIST_STORE(chip_array[chipid].watchp), &iter,
            WATCH_LOC, hex, WATCH_NAME, glob_name, WATCH_ADDR, addr, WATCH_SIZE, glob_size, 
            WATCH_ID, glob_id, -1 );
    atemu_add_watch(chipid, addr, glob_size);
    
    /* increment the watchpoint counter */
    chip_array[chipid].num_watches++;
    
    /* set watchpoint removal menu items to sensitive */
    watch_set_sensitive( TRUE, chipid );
    if(glob != NULL){
        watch_set_icon( TRUE, chipid, glob->iter );
    }
    
    select = gtk_tree_view_get_selection( GTK_TREE_VIEW( chip_array[chipid].treeview_watch ) );
    if( !gtk_tree_selection_get_selected( select, &tm, NULL ) )
    {
        gtk_tree_selection_select_iter( select, &iter );
    }
}

/* remove watchpoint */
/* id is optional.  
       if the caller knows which global this address corresponds to, set id
       otherwise, id = -1
 */
void remove_watchpoint(int chipid, long addr, int id)
{
    GtkTreeIter iter;
    long addrinnode;
    gboolean iterexists;
    int size;
    struct tuple_global_addr *glob;
    int global_id;

    iterexists = gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter );

    while( iterexists )
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter,
                WATCH_ADDR, &addrinnode, WATCH_SIZE, &size, WATCH_ID, &global_id, -1 );
        if ( ( addrinnode == addr ) && ( ( id == -1 ) || ( id == global_id ) ) )
        {
            glob = id_to_global_data(chipid, global_id);

            if(glob != NULL){
                watch_set_icon( FALSE, chipid, glob->iter );
            }
            
            move_selection( iter, GTK_TREE_VIEW( chip_array[chipid].treeview_watch ) );
            
            gtk_list_store_remove( GTK_LIST_STORE(chip_array[chipid].watchp),
                    &iter );
            break;
        }
        iterexists = gtk_tree_model_iter_next( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter );
    }

    /* decrement the watchpoint counter */
    chip_array[chipid].num_watches--;
    
    watch_set_sensitive( FALSE, chipid );
    
    atemu_clear_watch(chipid, addr);
}

/*test if it is a watchpoint*/
int is_watchpoint(int chipid, long addr)
{
    GtkTreeIter iter;
    long addrinnode;
    gboolean iterexists;

    iterexists = gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter );

    while( iterexists )
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter,
                WATCH_ADDR, &addrinnode, -1 );
        if( addrinnode == addr )
        {
            return 1;
        }
        iterexists = gtk_tree_model_iter_next( GTK_TREE_MODEL(chip_array[chipid].watchp), &iter );
    }
    return 0;
}
